import asyncio
import json
from abc import ABC, abstractmethod, abstractproperty
from copy import deepcopy

from .exceptions import (
    BotActionError,
    BotCheckerError,
    BotMatcherError,
    TryFlagFailed,
    get_better_exc,
)
from .typing import (
    TYPE_CHECKING,
    Any,
    BotLife,
    Callable,
    Coroutine,
    Enum,
    Literal,
    LogicMode,
    ModuleType,
    Optional,
    ParseArgs,
    Type,
    Void,
    VoidType,
)

if TYPE_CHECKING:
    from ..bot.hook import BotHookBus
    from ..controller.dispatcher import BotDispatcher
    from ..controller.responder import BotResponder
    from ..models.event import BotEventBuilder
    from ..plugin.handler import EventHandler
    from ..utils.logger import Logger


class AbstractConnector(ABC):
    """抽象连接器类

    .. admonition:: 提示
       :class: tip

       一般无需手动实例化该类，多数情况会直接使用本类对象，或将本类用作类型注解。
    """

    def __init__(self, cd_time: float) -> None:
        super().__init__()
        #: 连接器的日志器
        self.logger: "Logger"
        #: 是否在 slack 状态
        self.slack: bool = False
        #: 连接器发送行为操作的冷却时间
        self.cd_time = cd_time

        self._ref_flag: bool = False
        self._ready_signal = asyncio.Event()
        self._event_builder: Type["BotEventBuilder"]
        self._bot_bus: "BotHookBus"
        self._common_dispatcher: "BotDispatcher"
        self._resp_dispatcher: "BotResponder"

    @abstractmethod
    async def __aenter__(self):
        pass

    @abstractmethod
    async def __aexit__(
        self, exc_type: Type[Exception], exc_val: Exception, exc_tb: ModuleType
    ) -> bool:
        if exc_type is None:
            return True
        elif exc_type == asyncio.CancelledError:
            return True
        else:
            self.logger.error("连接器出现预期外的异常：\n" + get_better_exc(exc_val))
            return False

    def _set_ready(self) -> None:
        self._ready_signal.set()

    def _bind(
        self,
        dispatcher: "BotDispatcher",
        responder: "BotResponder",
        event_builder: Type["BotEventBuilder"],
        bot_bus: "BotHookBus",
        logger: "Logger",
    ) -> None:
        self._event_builder = event_builder
        self._bot_bus = bot_bus
        self.logger = logger
        self._common_dispatcher = dispatcher
        self._resp_dispatcher = responder

    @abstractmethod
    async def _alive_tasks(self) -> list[asyncio.Task]:
        pass

    @abstractmethod
    async def _send(self, action: "BotAction") -> None:
        pass


class Flagable:
    def __init__(self) -> None:
        self._flags_store: Optional[dict[str, dict[str, Any]]] = None

    def mark(self, namespace: str, flag_name: str, val: Any = None) -> None:
        """为对象添加在指定命名空间下，名为 flag_name 的标记。 此后此对象会一直携带此标记，无法撤销。"""
        if self._flags_store is None:
            self._flags_store = {}
        if self._flags_store.get(namespace) is None:
            self._flags_store[namespace] = {}
        if flag_name in self._flags_store[namespace].keys():
            raise TryFlagFailed(
                f"对象不可被重复标记。在命名空间 {namespace} 中名为 {flag_name} 的标记已存在"
            )
        self._flags_store[namespace][flag_name] = val

    def flag_check(self, namespace: str, flag_name: str, val: Any = None) -> bool:
        """检查此对象是否携带有指定的标记."""
        self._flags_store = self._flags_store
        if self._flags_store is None:
            return False
        if (flags := self._flags_store.get(namespace)) is None:
            return False
        if (flag := flags.get(flag_name, Void)) is Void:
            return False
        return flag is val if val is None else flag == val


class Cloneable:
    def copy(self):
        """返回一个本对象的一个深拷贝对象."""
        return deepcopy(self)


class BotEvent(ABC, Flagable):
    """事件基类

    .. admonition:: 提示
       :class: tip

       一般无需手动实例化该类，多数情况会直接使用本类对象，或将本类用作类型注解。
    """

    def __init__(self, rawEvent: dict) -> None:
        super().__init__()
        #: 从 onebot 实现项目获得的未格式化的事件
        self.raw = rawEvent
        self._args_map: Optional[dict[Any, dict[str, ParseArgs] | None]] = None

    @abstractproperty
    def time(self) -> int:
        """事件发生的时间"""
        pass

    @abstractproperty
    def type(self) -> Literal["message", "request", "notice", "meta", "response"]:
        """事件类型"""
        pass

    def is_msg_event(self) -> bool:
        """判断是否是消息事件"""
        return self.type == "message"

    def is_req_event(self) -> bool:
        """判断是否是请求事件"""
        return self.type == "request"

    def is_notice_event(self) -> bool:
        """判断是否是通知事件"""
        return self.type == "notice"

    def is_meta_event(self) -> bool:
        """判断是否是元事件"""
        return self.type == "meta"

    def is_resp_event(self) -> bool:
        """判断是否是响应事件"""
        return self.type == "response"

    def _get_args(self, parser_id: Any) -> dict[str, ParseArgs] | VoidType | None:
        if self._args_map is None:
            return Void
        return self._args_map.get(parser_id, Void)

    def _store_args(
        self, parser_id: Any, args_group: dict[str, ParseArgs] | None
    ) -> None:
        if self._args_map is None:
            self._args_map = {}
        self._args_map[parser_id] = args_group


class ActionArgs(ABC):
    # 行为信息构造基类

    def __init__(self) -> None:
        super().__init__()
        self.type: str
        self.params: dict


class BotAction(Flagable, Cloneable):
    """行为类

    每个行为操作应该产生一个行为对象

    .. admonition:: 提示
       :class: tip

       一般无需手动实例化该类，多数情况会直接使用本类对象，或将本类用作类型注解。
    """

    def __init__(
        self,
        action_args: ActionArgs,
        resp_id: Optional[str] = None,
        triggerEvent: Optional[BotEvent] = None,
        ready: bool = False,
    ) -> None:
        super().__init__()
        #: 响应标识 id。只有选择等待响应时，才会生成 id
        self.resp_id = resp_id
        #: 行为类型。对应 onebot 标准中的 API 终结点名称
        self.type = action_args.type
        #: 行为参数。对应 onebot 标准中向 API 传送的数据的 params 字段
        self.params = action_args.params
        #: 行为的触发事件，一般是行为生成时所在会话的事件
        self.trigger = triggerEvent

        self._ready = ready

    def extract(self) -> dict:
        # 从对象提取标准 cq action dict
        obj = {
            "action": self.type,
            "params": self.params,
        }
        if self.resp_id:
            obj["echo"] = self.resp_id
        return obj

    def flatten(self, indent: Optional[int] = None) -> str:
        # 将对象序列化为标准 cq action json 字符串，一般供连接器使用
        return json.dumps(self.extract(), ensure_ascii=False, indent=indent)

    def _fill_trigger(self, event: BotEvent) -> None:
        # 后期指定触发 event
        if self.trigger is None:
            self.trigger = event
            return
        raise BotActionError("action 已记录触发 event，拒绝再次记录")


class SessionRule(ABC):
    """会话规则基类

    会话规则用于两事件是否在同一会话的判断。
    """

    def __init__(self) -> None:
        super().__init__()

    @abstractmethod
    def compare(self, e1: BotEvent, e2: BotEvent) -> bool:
        """判断两事件是否在同一会话中的判断方法

        任何会话规则应该实现此抽象方法。

        :param e1: 判断时的事件1
        :param e2: 判断时的事件2
        :return: 判断结果
        """
        pass


class EventHandlerArgs:
    """事件方法（事件执行器）构造参数."""

    def __init__(
        self,
        executor: Callable[[], Coroutine[Any, Any, None]],
        type: Type["EventHandler"],
        params: list[Any],
    ) -> None:
        self.executor = executor
        self.type = type
        self.params = params


class ShareObjArgs:
    """插件共享对象构造参数."""

    def __init__(
        self, namespace: str, id: str, reflector: Callable[[], Coroutine[Any, Any, Any]]
    ) -> None:
        self.reflector = reflector
        self.namespace = namespace
        self.id = id


class ShareObjCbArgs:
    """插件共享对象回调的构造参数."""

    def __init__(
        self, namespace: str, id: str, cb: Callable[..., Coroutine[Any, Any, Any]]
    ) -> None:
        self.namespace = namespace
        self.id = id
        self.cb = cb


class PluginSignalHandlerArgs:
    """插件信号方法构造参数."""

    def __init__(
        self, func: Callable[..., Coroutine[Any, Any, Any]], namespace: str, signal: str
    ) -> None:
        self.func = func
        self.namespace = namespace
        self.signal = signal


class BotHookRunnerArgs:
    """钩子方法（生命周期回调）构造参数."""

    def __init__(
        self, func: Callable[..., Coroutine[Any, Any, None]], type: BotLife
    ) -> None:
        self.func = func
        self.type = type


class BotChecker(ABC, Cloneable):
    """检查器基类"""

    def __init__(
        self,
        ok_cb: Optional[Callable[[], Coroutine[Any, Any, None]]] = None,
        fail_cb: Optional[Callable[[], Coroutine[Any, Any, None]]] = None,
    ) -> None:
        """初始化一个检查器基类对象

        :param ok_cb: 检查通过的回调
        :param fail_cb: 检查不通过的回调
        """
        super().__init__()
        self.ok_cb = ok_cb
        self.fail_cb = fail_cb

    def __and__(self, other: "BotChecker") -> "WrappedChecker":
        if not isinstance(other, BotChecker):
            raise BotCheckerError(
                f"联合检查器定义时出现了非检查器对象，其值为：{other}"
            )
        return WrappedChecker(LogicMode.AND, self, other)

    def __or__(self, other: "BotChecker") -> "WrappedChecker":
        if not isinstance(other, BotChecker):
            raise BotCheckerError(
                f"联合检查器定义时出现了非检查器对象，其值为：{other}"
            )
        return WrappedChecker(LogicMode.OR, self, other)

    def __invert__(self) -> "WrappedChecker":
        return WrappedChecker(LogicMode.NOT, self)

    def __xor__(self, other: "BotChecker") -> "WrappedChecker":
        if not isinstance(other, BotChecker):
            raise BotCheckerError(
                f"联合检查器定义时出现了非检查器对象，其值为：{other}"
            )
        return WrappedChecker(LogicMode.XOR, self, other)

    def _fill_ok_cb(self, ok_cb: Callable[[], Coroutine[Any, Any, None]]) -> None:
        """后期指定 ok_cb 回调."""
        if self.ok_cb is not None:
            raise BotCheckerError(f"ok_cb 回调已经被初始化，值为：{self.ok_cb}")
        self.ok_cb = ok_cb

    def _fill_fail_cb(self, fail_cb: Callable[[], Coroutine[Any, Any, None]]) -> None:
        """后期指定 fail_cb 回调."""
        if self.fail_cb is not None:
            raise BotCheckerError(f"fail_cb 回调已经被初始化，值为：{self.fail_cb}")
        self.fail_cb = fail_cb

    @abstractmethod
    async def check(self, event: BotEvent) -> bool:
        """检查器检查方法

        任何检查器应该实现此抽象方法。

        :param event: 给定的事件
        :return: 检查是否通过
        """
        pass


class WrappedChecker(BotChecker):
    """合并检查器

    在两个 :class:`BotChecker` 对象间使用 | & ^ ~ 运算符即可返回合并检查器。

    .. admonition:: 提示
       :class: tip

       一般无需手动实例化该类，多数情况会直接使用本类对象，或将本类用作类型注解。
    """

    def __init__(
        self,
        mode: LogicMode,
        checker1: BotChecker,
        checker2: Optional[BotChecker] = None,
    ) -> None:
        """初始化一个合并检查器

        :param mode: 合并检查的逻辑模式
        :param checker1: 检查器1
        :param checker2: 检查器2
        """
        super().__init__()
        self.mode = mode
        self.c1, self.c2 = checker1, checker2

    def _fill_ok_cb(self, ok_cb: Callable[[], Coroutine[Any, Any, None]]) -> None:
        """后期指定 ok_cb 回调，注意此时是联合检查器， 因此将被自动应用到所包含的所有检查器."""
        super()._fill_ok_cb(ok_cb)
        self.c1._fill_ok_cb(ok_cb)
        self.c2._fill_ok_cb(ok_cb) if self.c2 else None

    def _fill_fail_cb(self, fail_cb: Callable[[], Coroutine[Any, Any, None]]) -> None:
        """后期指定 fail_cb 回调，注意此时是联合检查器， 因此将被自动应用到所包含的所有检查器."""
        super()._fill_fail_cb(fail_cb)
        self.c1._fill_fail_cb(fail_cb)
        self.c2._fill_fail_cb(fail_cb) if self.c2 else None

    async def check(self, event: BotEvent) -> bool:
        return LogicMode.calc(
            self.mode,
            await self.c1.check(event),
            (await self.c2.check(event)) if self.c2 is not None else None,
        )


class BotMatcher(ABC, Cloneable):
    """匹配器基类"""

    def __init__(self) -> None:
        """初始化一个检查器基类对象"""
        super().__init__()

    def __and__(self, other: "BotMatcher") -> "WrappedMatcher":
        if not isinstance(other, BotMatcher):
            raise BotMatcherError(
                f"联合匹配器定义时出现了非匹配器对象，其值为：{other}"
            )
        return WrappedMatcher(LogicMode.AND, self, other)

    def __or__(self, other: "BotMatcher") -> "WrappedMatcher":
        if not isinstance(other, BotMatcher):
            raise BotMatcherError(
                f"联合匹配器定义时出现了非匹配器对象，其值为：{other}"
            )
        return WrappedMatcher(LogicMode.OR, self, other)

    def __invert__(self) -> "WrappedMatcher":
        return WrappedMatcher(LogicMode.NOT, self)

    def __xor__(self, other: "BotMatcher") -> "WrappedMatcher":
        if not isinstance(other, BotMatcher):
            raise BotMatcherError(
                f"联合匹配器定义时出现了非匹配器对象，其值为：{other}"
            )
        return WrappedMatcher(LogicMode.XOR, self, other)

    @abstractmethod
    def match(self, text: str) -> bool:
        """匹配器匹配方法

        任何匹配器应该实现此抽象方法。

        :param text: 消息事件的文本内容
        :return: 是否匹配
        """
        pass


class WrappedMatcher(BotMatcher):
    """合并匹配器

    在两个 :class:`BotMatcher` 对象间使用 | & ^ ~ 运算符即可返回合并匹配器

    .. admonition:: 提示
       :class: tip

       一般无需手动实例化该类，多数情况会直接使用本类对象，或将本类用作类型注解。
    """

    def __init__(
        self,
        mode: LogicMode,
        matcher1: BotMatcher,
        matcher2: Optional[BotMatcher] = None,
    ) -> None:
        """初始化一个合并匹配器

        :param mode: 合并匹配的逻辑模式
        :param matcher1: 匹配器1
        :param matcher2: 匹配器2
        """
        super().__init__()
        self.mode = mode
        self.m1, self.m2 = matcher1, matcher2

    def match(self, text: str) -> bool:
        return LogicMode.calc(
            self.mode,
            self.m1.match(text),
            self.m2.match(text) if self.m2 is not None else None,
        )


class BotParser(ABC):
    # 解析器基类。解析器一般用作从消息文本中按规则提取指定字符串或字符串组合

    def __init__(self, id: Any) -> None:
        super().__init__()
        self.id = id
        self.need_format: bool = False

    @abstractmethod
    def parse(self, text: str) -> Optional[dict[str, ParseArgs]]:
        pass

    @abstractmethod
    def test(
        self, args_group: dict[str, ParseArgs] | None
    ) -> tuple[bool, Optional[str], Optional[ParseArgs]]:
        pass

    @abstractmethod
    async def format(self, cmd_name: str, args: ParseArgs) -> bool:
        pass



import asyncio

from ..base.exceptions import PluginSignalError, ShareObjError, get_better_exc
from ..base.tools import RWController, get_rich_str, to_task
from ..base.typing import TYPE_CHECKING, Any, Callable, Coroutine

if TYPE_CHECKING:
    from ..utils.logger import Logger


class ShareObject:
    """共享对象."""

    def __init__(self, namespace: str, id: str) -> None:
        self.namespace = namespace
        self.id = id
        self.__rwc = RWController()
        self.__reflect: Callable[[], Coroutine[Any, Any, Any]]
        self.__callback: Callable[..., Coroutine[Any, Any, Any]]

        self.__cb_set = asyncio.Event()

    def _fill_ref(self, reflector: Callable[[], Coroutine[Any, Any, Any]]) -> None:
        self.__reflect = reflector

    def _fill_cb(self, callback: Callable[..., Coroutine[Any, Any, Any]]) -> None:
        self.__callback = callback
        self.__cb_set.set()

    @property
    async def val(self) -> Any:
        """共享对象引用的值."""
        async with self.__rwc.safe_read():
            return await self.__reflect()

    async def affect(self, *args: Any, **kwargs: Any) -> Any:
        """触发共享对象的回调，回调未设置时会等待。 如果本来就没有回调，则会陷入无休止等待."""
        await self.__cb_set.wait()
        async with self.__rwc.safe_write():
            return await self.__callback(*args, **kwargs)


class PluginStore:
    """插件共享存储."""

    def __init__(self) -> None:
        self.store: dict[str, dict[str, ShareObject]] = {}

    def create_so(
        self, reflector: Callable[[], Coroutine[Any, Any, Any]], namespace: str, id: str
    ) -> None:
        """创建共享对象."""
        if namespace not in self.store.keys():
            self.store[namespace] = {}
        obj = self.store[namespace].get(id)
        if obj is not None:
            raise ShareObjError(
                f"已在 {namespace} 命名空间中注册标记为 {id} 的共享对象，拒绝再次注册"
            )
        obj = ShareObject(namespace, id)
        self.store[namespace][id] = obj
        obj._fill_ref(reflector)

    def bind_cb(
        self, namespace: str, id: str, cb: Callable[..., Coroutine[Any, Any, Any]]
    ) -> None:
        """为共享对象绑定回调."""
        if namespace not in self.store.keys():
            raise ShareObjError(f"共享对象回调指定的命名空间 {namespace} 不存在")
        if id not in self.store[namespace].keys():
            raise ShareObjError(
                f"共享对象回调指定的命名空间中，不存在标记为 {id} 的共享对象"
            )
        if self.store[namespace][id].__cb_set.is_set():
            raise ShareObjError(
                f"{namespace} 中标记为 {id} 的共享对象已被注册过回调，拒绝再次注册"
            )
        self.store[namespace][id]._fill_cb(cb)

    def get(self, namespace: str, id: str) -> ShareObject:
        """获取共享对象."""
        if namespace not in self.store.keys():
            raise ShareObjError(
                f"无法获取不存在的共享对象：命名空间 {namespace} 不存在"
            )
        if id not in self.store[namespace].keys():
            raise ShareObjError(f"无法获取不存在的共享对象：标识 {id} 不存在")
        return self.store[namespace][id]


class PluginSignalHandler:
    """插件信号处理器."""

    def __init__(
        self, namespace: str, signal: str, func: Callable[..., Coroutine[Any, Any, Any]]
    ) -> None:
        self.cb = func
        self.namespace = namespace
        self.signal = signal


class PluginBus:
    """插件信号总线."""

    def __init__(self) -> None:
        self.store: dict[str, dict[str, PluginSignalHandler]] = {}
        self.logger: "Logger"

    def _bind(self, logger: "Logger") -> None:
        self.logger = logger

    def register(
        self, namespace: str, signal: str, func: Callable[..., Coroutine[Any, Any, Any]]
    ) -> None:
        """注册一个插件信号处理方法。由 plugin build 过程调用."""
        if namespace not in self.store.keys():
            self.store[namespace] = {}
        if signal in self.store[namespace].keys():
            raise PluginSignalError("同一命名空间的同一信号只能绑定一个处理函数")
        self.store[namespace][signal] = PluginSignalHandler(namespace, signal, func)

    async def _run_on_ctx(
        self, handler: PluginSignalHandler, *args: Any, **kwargs: Any
    ) -> Any:
        """在指定的上下文下运行插件信号处理方法."""
        try:
            ret = await handler.cb(*args, **kwargs)
            return ret
        except Exception as e:
            func_name = handler.cb.__qualname__
            self.logger.error(f"插件信号处理方法 {func_name} 发生异常")
            self.logger.error("异常回溯栈：\n" + get_better_exc(e))
            self.logger.error("异常点局部变量：\n" + get_rich_str(locals()))

    async def emit(
        self, namespace: str, signal: str, *args: Any, wait: bool = False, **kwargs: Any
    ) -> Any:
        """触发一个插件信号。如果指定 wait 为 True，则会等待所有插件信号处理方法完成。 若启用 forward，则会将 session
        从信号触发处转发到信号处理处。 但启用 forward，必须同时启用 wait。

        注意：如果你要等待返回结果，则需要指定 wait=True，否则不会等待且始终返回 None
        """
        if namespace not in self.store.keys():
            raise PluginSignalError(
                f"插件信号命名空间 {namespace} 不存在，无法触发信号"
            )
        if signal not in self.store[namespace].keys():
            self.logger.warning(
                f"命名空间 {namespace} 内，没有处理信号 {signal} 的处理函数"
            )
            return

        handler = self.store[namespace][signal]
        if not wait:
            to_task(self._run_on_ctx(handler, *args, **kwargs))
            return None
        else:
            return await self._run_on_ctx(handler, *args, **kwargs)



import asyncio
import time
from itertools import count

import websockets
import websockets.exceptions as wse

from ..base.abc import AbstractConnector, BotLife
from ..base.exceptions import BotConnectFailed, get_better_exc
from ..base.tools import get_rich_str, to_task
from ..base.typing import TYPE_CHECKING, ModuleType, Type

if TYPE_CHECKING:
    import websockets.client

    from ..base.abc import BotAction


class ForwardWsConn(AbstractConnector):
    """正向 websocket 连接器

    .. admonition:: 注意
       :class: caution

       注意：在 melobot 中，正向 websocket 连接器会开启一个 ws 客户端。这个客户端只能和一个服务端通信。

    正向 websocket 通信方式暂时不支持断连后尝试重连。断连后将会直接停止 bot
    """

    def __init__(
        self,
        connect_host: str,
        connect_port: int,
        max_retry: int = -1,
        retry_delay: float = 5.0,
        cd_time: float = 0.2,
    ) -> None:
        """初始化一个正向 websocket 连接器

        :param connect_host: 连接的 host
        :param connect_port: 连接的 port
        :param max_retry: 初始连接最大重试次数，默认 -1 代表无限次重试
        :param retry_delay: 初始连接重试间隔时间
        :param cd_time: 行为操作冷却时间（用于防止风控）
        """
        super().__init__(cd_time)
        #: 连接失败最大重试次数
        self.max_retry: int = max_retry
        #: 连接失败重试间隔
        self.retry_delay: float = retry_delay if retry_delay > 0 else 0
        #: ws 连接的 url（形如：ws://xxx:xxx）
        self.url = f"ws://{connect_host}:{connect_port}"
        #: 连接对象
        self.conn: "websockets.client.WebSocketClientProtocol"

        self._send_queue: asyncio.Queue["BotAction"] = asyncio.Queue()
        self._pre_send_time = time.time()

    async def _start(self) -> None:
        """启动连接."""
        iterator = count(0) if self.max_retry < 0 else range(self.max_retry + 1)
        for _ in iterator:
            try:
                self.conn = await websockets.connect(self.url)
                await self.conn.recv()

                self.logger.info("连接器与前端建立了 ws 连接")
                await self._bot_bus.emit(BotLife.CONNECTED)
                self.logger.debug("CONNECTED hook 已完成")
                return
            except Exception as e:
                self.logger.warning(
                    f"ws 连接建立失败，{self.retry_delay}s 后自动重试。错误：{e}"
                )
                await asyncio.sleep(self.retry_delay)
        raise BotConnectFailed("连接重试已达最大重试次数，已放弃建立连接")

    async def _close(self) -> None:
        """关闭连接."""
        await self.conn.close()
        await self.conn.wait_closed()
        self.logger.info("连接器与前端的连接已安全关闭")

    async def __aenter__(self) -> "ForwardWsConn":
        await self._start()
        return self

    async def __aexit__(
        self, exc_type: Type[Exception], exc_val: Exception, exc_tb: ModuleType
    ) -> bool:
        await self._close()
        return await super().__aexit__(exc_type, exc_val, exc_tb)

    async def _send(self, action: "BotAction") -> None:
        """发送一个 action 给连接器。实际上是先提交到 send_queue."""
        await self._ready_signal.wait()

        if self.slack:
            self.logger.debug(f"action {id(action)} 因 slack 状态被丢弃")
            return
        await self._send_queue.put(action)
        self.logger.debug(f"action {id(action)} 已成功加入发送队列")

    async def _send_queue_watch(self) -> None:
        """真正的发送方法。从 send_queue 提取 action 并按照一些处理步骤操作."""
        await self._ready_signal.wait()

        try:
            while True:
                action = await self._send_queue.get()
                self.logger.debug(
                    f"action {id(action)} 准备发送，结构如下：\n"
                    + get_rich_str(action.__dict__)
                )
                await self._bot_bus.emit(BotLife.ACTION_PRESEND, action, wait=True)
                self.logger.debug(f"action {id(action)} presend hook 已完成")
                action_str = action.flatten()
                wait_time = self.cd_time - (time.time() - self._pre_send_time)
                self.logger.debug(f"action {id(action)} 冷却等待：{wait_time}")
                await asyncio.sleep(wait_time)
                await self.conn.send(action_str)
                self.logger.debug(f"action {id(action)} 已发送")
                self._pre_send_time = time.time()
        except asyncio.CancelledError:
            self.logger.debug("连接器发送队列监视任务已被结束")
        except wse.ConnectionClosed:
            self.logger.error("连接器与前端的通信已经停止，无法再执行行为操作")

    async def _listen(self) -> None:
        """从前端接收一个事件，并处理"""
        await self._ready_signal.wait()

        try:
            while True:
                try:
                    raw_event = await self.conn.recv()
                    self.logger.debug(f"收到事件，未格式化的字符串：\n{raw_event}")
                    if raw_event == "":
                        continue
                    event = self._event_builder.build(raw_event)
                    self.logger.debug(
                        f"event {id(event)} 构建完成，结构：\n"
                        + get_rich_str(event.raw)
                    )
                    if event.is_resp_event():
                        to_task(self._resp_dispatcher.respond(event))  # type: ignore
                    else:
                        to_task(self._common_dispatcher.dispatch(event))  # type: ignore
                except wse.ConnectionClosed:
                    raise
                except Exception as e:
                    self.logger.error("bot 连接器监听任务抛出异常")
                    self.logger.error(f"异常点 raw_event：{raw_event}")
                    self.logger.error("异常回溯栈：\n" + get_better_exc(e))
                    self.logger.error("异常点局部变量：\n" + get_rich_str(locals()))
        except asyncio.CancelledError:
            self.logger.debug("连接器监听任务已停止")
        except wse.ConnectionClosed:
            self.logger.debug("连接器与前端的通信已经停止")

    async def _alive_tasks(self) -> list[asyncio.Task]:
        to_task(self._send_queue_watch())
        return [to_task(self._listen())]



import logging
import logging.config
import logging.handlers
import os
from collections.abc import Mapping
from logging import CRITICAL, DEBUG, ERROR, INFO, WARNING, Logger
from types import TracebackType

import coloredlogs

from ..base.exceptions import DuplicateError
from ..base.typing import Literal, Optional, Type, TypeAlias


class NullLogger(Logger):
    """获得一个空日志器，支持所有日志操作， 但是丢弃所有日志."""

    def __init__(self, name: str) -> None:
        super().__init__(name, CRITICAL)
        self.addHandler(logging.NullHandler())


class BotLogger(Logger):
    """日志器类"""

    LOGGERS: dict[str, "BotLogger"] = {}

    FIELD_COLORS = {
        "asctime": {"color": "white"},
        "hostname": {"color": "magenta"},
        "levelname": {"bold": True, "color": "white"},
        "name": {"color": "blue"},
        "programname": {"color": "cyan"},
        "username": {"color": "yellow"},
    }

    LEVEL_COLORS = {
        "critical": {"bold": True, "color": "red"},
        "debug": {"color": "magenta"},
        "error": {"color": "red"},
        "info": {},
        "notice": {"color": "magenta"},
        "spam": {"color": "green", "faint": True},
        "success": {"bold": True, "color": "green"},
        "verbose": {"color": "blue"},
        "warning": {"color": "yellow"},
    }

    LEVEL_MAP = {
        "DEBUG": DEBUG,
        "INFO": INFO,
        "WARNING": WARNING,
        "ERROR": ERROR,
        "CRITICAL": CRITICAL,
    }

    @staticmethod
    def _console_fmt(name: str, no_tag: bool = False) -> logging.Formatter:
        fmt_s = (
            f"[%(asctime)s] [%(levelname)s] [{name}] %(message)s"
            if not no_tag
            else "[%(asctime)s] [%(levelname)s] %(message)s"
        )
        fmt = coloredlogs.ColoredFormatter(
            fmt=fmt_s,
            datefmt="%Y-%m-%d %H:%M:%S",
            level_styles=BotLogger.LEVEL_COLORS,
            field_styles=BotLogger.FIELD_COLORS,
        )
        return fmt

    @staticmethod
    def _console_handler(fmt: logging.Formatter) -> logging.Handler:
        handler = logging.StreamHandler()
        handler.setFormatter(fmt)
        return handler

    @staticmethod
    def _file_fmt(name: str, no_tag: bool = False) -> logging.Formatter:
        fmt_s = (
            f"[%(asctime)s] [%(filename)s %(lineno)d] [%(levelname)s] [{name}] %(message)s"
            if not no_tag
            else "[%(asctime)s] [%(filename)s %(lineno)d] [%(levelname)s] %(message)s"
        )
        fmt = logging.Formatter(
            fmt=fmt_s,
            datefmt="%Y-%m-%d %H:%M:%S",
        )
        return fmt

    @staticmethod
    def _file_handler(
        fmt: logging.Formatter, log_dir: str, name: str
    ) -> logging.Handler:
        if not os.path.exists(log_dir):
            os.mkdir(log_dir)
        handler = logging.handlers.RotatingFileHandler(
            filename=os.path.join(log_dir, f"{name}.log"),
            maxBytes=1024 * 1024,
            backupCount=10,
            encoding="UTF-8",
        )
        handler.setFormatter(fmt)
        return handler

    def __init__(
        self,
        name: str,
        level: Literal["DEBUG", "ERROR", "INFO", "WARNING", "CRITICAL"] = "INFO",
        to_console: bool = True,
        to_dir: Optional[str] = None,
        no_tag: bool = False,
    ) -> None:
        """初始化一个日志器实例

        :param name: 日志器的名称（唯一）
        :param level: 日志等级
        :param to_console: 是否输出到 console
        :param to_dir: 保存日志文件的目录，为空则不保存文件
        :param no_tag: 记录日志时是否不标识日志器名称
        """

        if name in BotLogger.LOGGERS.keys():
            raise DuplicateError(f"名为 {name} 的日志器已存在，请修改 name")
        super().__init__(name, BotLogger.LEVEL_MAP[level])
        BotLogger.LOGGERS[name] = self
        self._con_handler: Optional[logging.Handler] = None
        self._handler_arr: list[logging.Handler] = []
        self._no_tag = no_tag

        if to_console:
            self._add_console_handler()
        if to_dir is not None:
            self._add_file_handler(to_dir)

    def _add_console_handler(self) -> None:
        if self._con_handler is None:
            self._con_handler = self._console_handler(
                self._console_fmt(self.name, self._no_tag)
            )
            self.addHandler(self._con_handler)
            self._handler_arr.append(self._con_handler)

    def _add_file_handler(self, log_dir: str) -> None:
        handler = self._file_handler(
            self._file_fmt(self.name, self._no_tag), log_dir, self.name
        )
        self.addHandler(handler)
        self._handler_arr.append(handler)

    def setLevel(self, level: Literal["DEBUG", "ERROR", "INFO", "WARNING", "CRITICAL"]) -> None:  # type: ignore
        super().setLevel(level)
        for handler in self._handler_arr:
            handler.setLevel(level)

    # def to_console(self) -> None:
    #     self._add_console_handler()

    # def to_dir(self, log_dir: str) -> None:
    #     self._add_file_handler(log_dir)


_SysExcInfoType: TypeAlias = (
    tuple[Type[BaseException], BaseException, Optional[TracebackType]]
    | tuple[None, None, None]
)
_ExcInfoType: TypeAlias = None | bool | _SysExcInfoType | BaseException


class PrefixLogger:
    """二次包装的日志器."""

    def __init__(self, ref: BotLogger, prefix: str) -> None:
        self._logger = ref
        self._prefix = prefix

    def _add_prefix(self, s: object) -> str:
        return f"[{self._prefix}] {s}"

    def info(
        self,
        msg: object,
        *args: object,
        exc_info: _ExcInfoType = None,
        stack_info: bool = False,
        stacklevel: int = 1,
        extra: Optional[Mapping[str, object]] = None,
    ) -> None:
        msg = self._add_prefix(msg)
        return self._logger.info(
            msg,
            *args,
            exc_info=exc_info,
            stack_info=stack_info,
            stacklevel=stacklevel,
            extra=extra,
        )

    def warning(
        self,
        msg: object,
        *args: object,
        exc_info: _ExcInfoType = None,
        stack_info: bool = False,
        stacklevel: int = 1,
        extra: Optional[Mapping[str, object]] = None,
    ) -> None:
        msg = self._add_prefix(msg)
        return self._logger.warning(
            msg,
            *args,
            exc_info=exc_info,
            stack_info=stack_info,
            stacklevel=stacklevel,
            extra=extra,
        )

    def error(
        self,
        msg: object,
        *args: object,
        exc_info: _ExcInfoType = None,
        stack_info: bool = False,
        stacklevel: int = 1,
        extra: Optional[Mapping[str, object]] = None,
    ) -> None:
        msg = self._add_prefix(msg)
        return self._logger.error(
            msg,
            *args,
            exc_info=exc_info,
            stack_info=stack_info,
            stacklevel=stacklevel,
            extra=extra,
        )

    def debug(
        self,
        msg: object,
        *args: object,
        exc_info: _ExcInfoType = None,
        stack_info: bool = False,
        stacklevel: int = 1,
        extra: Optional[Mapping[str, object]] = None,
    ) -> None:
        msg = self._add_prefix(msg)
        return self._logger.debug(
            msg,
            *args,
            exc_info=exc_info,
            stack_info=stack_info,
            stacklevel=stacklevel,
            extra=extra,
        )

    def critical(
        self,
        msg: object,
        *args: object,
        exc_info: _ExcInfoType = None,
        stack_info: bool = False,
        stacklevel: int = 1,
        extra: Optional[Mapping[str, object]] = None,
    ) -> None:
        msg = self._add_prefix(msg)
        return self._logger.critical(
            msg,
            *args,
            exc_info=exc_info,
            stack_info=stack_info,
            stacklevel=stacklevel,
            extra=extra,
        )



import importlib.util
import os
import pathlib
import sys

from ..base.abc import (
    BotHookRunnerArgs,
    BotLife,
    EventHandlerArgs,
    PluginSignalHandlerArgs,
    ShareObjArgs,
    ShareObjCbArgs,
)
from ..base.exceptions import PluginInitError
from ..base.tools import to_async
from ..base.typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Coroutine,
    Literal,
    LogicMode,
    Optional,
    P,
    PriorLevel,
    Union,
)
from ..utils.checker import (
    AtChecker,
    FriendReqChecker,
    GroupReqChecker,
    NoticeTypeChecker,
)
from ..utils.matcher import (
    ContainMatcher,
    EndMatcher,
    FullMatcher,
    RegexMatcher,
    StartMatcher,
)
from .handler import (
    AllEventHandler,
    MetaEventHandler,
    MsgEventHandler,
    NoticeEventHandler,
    ReqEventHandler,
)

if TYPE_CHECKING:
    from ..base.abc import SessionRule, WrappedChecker
    from ..utils.checker import BotChecker
    from ..utils.matcher import BotMatcher
    from ..utils.parser import BotParser


class PluginProxy:
    """Bot 插件代理类。供外部使用."""

    def __init__(
        self,
        id: str,
        ver: str,
        desc: str,
        doc: str,
        keywords: list[str],
        url: str,
        share_objs: list[tuple[str, str]],
        share_cbs: list[tuple[str, str]],
        signal_methods: list[tuple[str, str]],
    ) -> None:
        self.id = id
        self.version = ver
        self.desc = desc
        self.doc = doc
        self.keywords = keywords
        self.url = url
        self.shares = share_objs
        self.share_cbs = share_cbs
        self.signal_methods = signal_methods


class PluginLoader:
    """插件加载器."""

    @staticmethod
    def load_from_dir(plugin_path: str) -> "BotPlugin":
        """从指定插件目录加载插件."""
        if not os.path.exists(os.path.join(plugin_path, "__init__.py")):
            raise PluginInitError(
                f"{plugin_path} 缺乏入口主文件 __init__.py，插件无法加载"
            )
        plugin_name = os.path.basename(plugin_path)
        plugins_folder = str(pathlib.Path(plugin_path).parent.resolve(strict=True))
        plugins_folder_name = os.path.basename(plugins_folder)
        if plugins_folder not in sys.path:
            importlib.import_module(plugins_folder_name)
            sys.path.insert(0, plugins_folder)
        module = importlib.import_module(
            f"{plugins_folder_name}.{plugin_name}", f"{plugins_folder_name}"
        )

        plugin = None
        for obj in module.__dict__.values():
            if isinstance(obj, BotPlugin):
                plugin = obj
                break
        if plugin is None:
            raise PluginInitError("指定的入口主文件中，未发现 Plugin 实例，无效导入")
        return plugin

    @staticmethod
    def load(target: Union[str, "BotPlugin"]) -> "BotPlugin":
        """加载插件."""
        if isinstance(target, str):
            plugin = PluginLoader.load_from_dir(target)
        else:
            plugin = target
        plugin._self_build()
        return plugin


class BotPlugin:
    """插件类，使用该类实例化一个插件

    同时该类会提供各种接口用于注册事件处理器、共享对象、共享对象回调方法和信号处理方法。
    """

    def __init__(
        self,
        id: str,
        version: str,
        desc: str = "",
        doc: str = "",
        keywords: Optional[list[str]] = None,
        url: str = "",
    ) -> None:
        """初始化一个插件

        :param id: 插件的 id
        :param version: 插件的版本
        :param desc: 插件功能描述
        :param doc: 插件简单的文档说明
        :param keywords: 关键词列表
        :param url: 插件项目地址
        """
        self.__id__ = id
        self.__version__ = version
        self.__desc__ = desc
        self.__keywords__ = keywords if keywords is not None else []
        self.__url__ = url
        self.__pdoc__ = doc

        self.__handler_args__: list[EventHandlerArgs] = []
        self.__signal_args__: list[PluginSignalHandlerArgs] = []
        self.__share_args__: list[ShareObjArgs] = []
        self.__share_cb_args__: list[ShareObjCbArgs] = []
        self.__hook_args__: list[BotHookRunnerArgs] = []

        self.__proxy__: PluginProxy

    def _self_build(self) -> None:
        self.__proxy__ = PluginProxy(
            self.__id__,
            self.__version__,
            self.__desc__,
            self.__pdoc__,
            self.__keywords__,
            self.__url__,
            [(args.namespace, args.id) for args in self.__share_args__],
            [(args.namespace, args.id) for args in self.__share_cb_args__],
            [(args.namespace, args.signal) for args in self.__signal_args__],
        )
        check_pass = all(
            False
            for pair in self.__proxy__.share_cbs
            if pair not in self.__proxy__.shares
        )
        if not check_pass:
            raise PluginInitError(
                f"插件 {self.__id__} 不能为不属于自己的共享对象注册回调"
            )

    def on_event(
        self,
        checker: Optional["BotChecker"] = None,
        priority: PriorLevel = PriorLevel.MEAN,
        block: bool = False,
        temp: bool = False,
        session_rule: Optional["SessionRule"] = None,
        session_hold: bool = False,
        direct_rouse: bool = False,
        conflict_wait: bool = False,
        conflict_cb: Optional[Callable[[], Coroutine[Any, Any, None]]] = None,
    ):
        """注册一个任意事件处理器

        :param checker: 使用的检查器，为空则默认通过检查
        :param priority: 优先级
        :param block: 是否进行优先级阻断
        :param temp: 是否是一次性的
        :param session_rule: 会话规则，为空则不使用会话规则
        :param session_hold: 处理方法结束后是否保留会话（有会话规则才可启用）
        :param direct_rouse: 会话暂停时，是否允许不检查就唤醒会话（有会话规则才可启用）
        :param conflict_wait: 会话冲突时，是否需要事件等待处理（有会话规则才可启用）
        :param conflict_cb: 会话冲突时，运行的回调（有会话规则才可启用，`conflict_wait=True`，此参数无效）
        """

        def make_args(
            executor: Callable[[], Coroutine[Any, Any, None]]
        ) -> Callable[[], Coroutine[Any, Any, None]]:
            self.__handler_args__.append(
                EventHandlerArgs(
                    executor=executor,
                    type=AllEventHandler,
                    params=[
                        checker,
                        priority,
                        block,
                        temp,
                        session_rule,
                        session_hold,
                        direct_rouse,
                        conflict_wait,
                        conflict_cb,
                    ],
                )
            )
            return executor

        return make_args

    def on_message(
        self,
        matcher: Optional["BotMatcher"] = None,
        parser: Optional["BotParser"] = None,
        checker: Optional["BotChecker"] = None,
        priority: PriorLevel = PriorLevel.MEAN,
        block: bool = False,
        temp: bool = False,
        session_rule: Optional["SessionRule"] = None,
        session_hold: bool = False,
        direct_rouse: bool = False,
        conflict_wait: bool = False,
        conflict_cb: Optional[Callable[[], Coroutine[Any, Any, None]]] = None,
    ):
        """注册一个消息事件处理器

        :param matcher: 使用的匹配器（和解析器二选一）
        :param parser: 使用的解析器（和匹配器二选一）
        :param checker: 使用的检查器，为空则默认通过检查
        :param priority: 优先级
        :param block: 是否进行优先级阻断
        :param temp: 是否是一次性的
        :param session_rule: 会话规则，为空则不使用会话规则
        :param session_hold: 处理方法结束后是否保留会话（有会话规则才可启用）
        :param direct_rouse: 会话暂停时，是否允许不检查就唤醒会话（有会话规则才可启用）
        :param conflict_wait: 会话冲突时，是否需要事件等待处理（有会话规则才可启用）
        :param conflict_cb: 会话冲突时，运行的回调（有会话规则才可启用，`conflict_wait=True`，此参数无效）
        """

        def make_args(
            executor: Callable[[], Coroutine[Any, Any, None]]
        ) -> Callable[[], Coroutine[Any, Any, None]]:
            self.__handler_args__.append(
                EventHandlerArgs(
                    executor=executor,
                    type=MsgEventHandler,
                    params=[
                        matcher,
                        parser,
                        checker,
                        priority,
                        block,
                        temp,
                        session_rule,
                        session_hold,
                        direct_rouse,
                        conflict_wait,
                        conflict_cb,
                    ],
                )
            )
            return executor

        return make_args

    def on_every_message(
        self,
        checker: Optional["BotChecker"] = None,
        priority: PriorLevel = PriorLevel.MEAN,
        block: bool = False,
        temp: bool = False,
        session_rule: Optional["SessionRule"] = None,
        session_hold: bool = False,
        direct_rouse: bool = False,
        conflict_wait: bool = False,
        conflict_cb: Optional[Callable[[], Coroutine[Any, Any, None]]] = None,
    ):
        """注册一个任意消息事件处理器

        :param checker: 使用的检查器，为空则默认通过检查
        :param priority: 优先级
        :param block: 是否进行优先级阻断
        :param temp: 是否是一次性的
        :param session_rule: 会话规则，为空则不使用会话规则
        :param session_hold: 处理方法结束后是否保留会话（有会话规则才可启用）
        :param direct_rouse: 会话暂停时，是否允许不检查就唤醒会话（有会话规则才可启用）
        :param conflict_wait: 会话冲突时，是否需要事件等待处理（有会话规则才可启用）
        :param conflict_cb: 会话冲突时，运行的回调（有会话规则才可启用，`conflict_wait=True`，此参数无效）
        """

        def make_args(
            executor: Callable[[], Coroutine[Any, Any, None]]
        ) -> Callable[[], Coroutine[Any, Any, None]]:
            self.__handler_args__.append(
                EventHandlerArgs(
                    executor=executor,
                    type=MsgEventHandler,
                    params=[
                        None,
                        None,
                        checker,
                        priority,
                        block,
                        temp,
                        session_rule,
                        session_hold,
                        direct_rouse,
                        conflict_wait,
                        conflict_cb,
                    ],
                )
            )
            return executor

        return make_args

    def on_at_qq(
        self,
        qid: Optional[int] = None,
        matcher: Optional["BotMatcher"] = None,
        parser: Optional["BotParser"] = None,
        checker: Optional["BotChecker"] = None,
        priority: PriorLevel = PriorLevel.MEAN,
        block: bool = False,
        temp: bool = False,
        session_rule: Optional["SessionRule"] = None,
        session_hold: bool = False,
        direct_rouse: bool = False,
        conflict_wait: bool = False,
        conflict_cb: Optional[Callable[[], Coroutine[Any, Any, None]]] = None,
    ):
        """注册一个艾特消息事件处理器

        消息必须是艾特消息，且匹配成功才能被进一步处理。

        :param qid: 被艾特的 qq 号。为空则接受所有艾特消息;不为空则只接受指定 qid 被艾特的艾特消息
        :param matcher: 使用的匹配器（和解析器二选一）
        :param parser: 使用的解析器（和匹配器二选一）
        :param checker: 使用的检查器，为空则默认通过检查
        :param priority: 优先级
        :param block: 是否进行优先级阻断
        :param temp: 是否是一次性的
        :param session_rule: 会话规则，为空则不使用会话规则
        :param session_hold: 处理方法结束后是否保留会话（有会话规则才可启用）
        :param direct_rouse: 会话暂停时，是否允许不检查就唤醒会话（有会话规则才可启用）
        :param conflict_wait: 会话冲突时，是否需要事件等待处理（有会话规则才可启用）
        :param conflict_cb: 会话冲突时，运行的回调（有会话规则才可启用，`conflict_wait=True`，此参数无效）
        """
        at_checker = AtChecker(qid)
        wrapped_checker: AtChecker | "WrappedChecker"
        if checker is not None:
            wrapped_checker = at_checker & checker
        else:
            wrapped_checker = at_checker

        def make_args(
            executor: Callable[[], Coroutine[Any, Any, None]]
        ) -> Callable[[], Coroutine[Any, Any, None]]:
            self.__handler_args__.append(
                EventHandlerArgs(
                    executor=executor,
                    type=MsgEventHandler,
                    params=[
                        matcher,
                        parser,
                        wrapped_checker,
                        priority,
                        block,
                        temp,
                        session_rule,
                        session_hold,
                        direct_rouse,
                        conflict_wait,
                        conflict_cb,
                    ],
                )
            )
            return executor

        return make_args

    def on_start_match(
        self,
        target: str | list[str],
        logic_mode: LogicMode = LogicMode.OR,
        checker: Optional["BotChecker"] = None,
        priority: PriorLevel = PriorLevel.MEAN,
        block: bool = False,
        temp: bool = False,
        session_rule: Optional["SessionRule"] = None,
        session_hold: bool = False,
        direct_rouse: bool = False,
        conflict_wait: bool = False,
        conflict_cb: Optional[Callable[[], Coroutine[Any, Any, None]]] = None,
    ):
        """注册一个字符串起始匹配的消息事件处理器

        `target` 为字符串时，只进行一次起始匹配，即判断是否匹配成功。
        `target` 为字符串列表时，所有字符串都进行起始匹配，再将所有结果使用给定
        `logic_mode` 计算是否匹配成功。

        消息必须匹配成功才能被进一步处理。

        :param target: 匹配目标
        :param logic_mode: `target` 为 `list[str]` 时的计算模式
        :param checker: 使用的检查器，为空则默认通过检查
        :param priority: 优先级
        :param block: 是否进行优先级阻断
        :param temp: 是否是一次性的
        :param session_rule: 会话规则，为空则不使用会话规则
        :param session_hold: 处理方法结束后是否保留会话（有会话规则才可启用）
        :param direct_rouse: 会话暂停时，是否允许不检查就唤醒会话（有会话规则才可启用）
        :param conflict_wait: 会话冲突时，是否需要事件等待处理（有会话规则才可启用）
        :param conflict_cb: 会话冲突时，运行的回调（有会话规则才可启用，`conflict_wait=True`，此参数无效）
        """
        start_matcher = StartMatcher(target, logic_mode)

        def make_args(
            executor: Callable[[], Coroutine[Any, Any, None]]
        ) -> Callable[[], Coroutine[Any, Any, None]]:
            self.__handler_args__.append(
                EventHandlerArgs(
                    executor=executor,
                    type=MsgEventHandler,
                    params=[
                        start_matcher,
                        None,
                        checker,
                        priority,
                        block,
                        temp,
                        session_rule,
                        session_hold,
                        direct_rouse,
                        conflict_wait,
                        conflict_cb,
                    ],
                )
            )
            return executor

        return make_args

    def on_contain_match(
        self,
        target: str | list[str],
        logic_mode: LogicMode = LogicMode.OR,
        checker: Optional["BotChecker"] = None,
        priority: PriorLevel = PriorLevel.MEAN,
        block: bool = False,
        temp: bool = False,
        session_rule: Optional["SessionRule"] = None,
        session_hold: bool = False,
        direct_rouse: bool = False,
        conflict_wait: bool = False,
        conflict_cb: Optional[Callable[[], Coroutine[Any, Any, None]]] = None,
    ):
        """注册一个字符串包含匹配的消息事件处理器

        `target` 为字符串时，只进行一次包含匹配，即判断是否匹配成功。
        `target` 为字符串列表时，所有字符串都进行包含匹配，再将所有结果使用给定
        `logic_mode` 计算是否匹配成功。

        消息必须匹配成功才能被进一步处理。

        :param target: 匹配目标
        :param logic_mode: `target` 为 `list[str]` 时的计算模式
        :param checker: 使用的检查器，为空则默认通过检查
        :param priority: 优先级
        :param block: 是否进行优先级阻断
        :param temp: 是否是一次性的
        :param session_rule: 会话规则，为空则不使用会话规则
        :param session_hold: 处理方法结束后是否保留会话（有会话规则才可启用）
        :param direct_rouse: 会话暂停时，是否允许不检查就唤醒会话（有会话规则才可启用）
        :param conflict_wait: 会话冲突时，是否需要事件等待处理（有会话规则才可启用）
        :param conflict_cb: 会话冲突时，运行的回调（有会话规则才可启用，`conflict_wait=True`，此参数无效）
        """
        contain_matcher = ContainMatcher(target, logic_mode)

        def make_args(
            executor: Callable[[], Coroutine[Any, Any, None]]
        ) -> Callable[[], Coroutine[Any, Any, None]]:
            self.__handler_args__.append(
                EventHandlerArgs(
                    executor=executor,
                    type=MsgEventHandler,
                    params=[
                        contain_matcher,
                        None,
                        checker,
                        priority,
                        block,
                        temp,
                        session_rule,
                        session_hold,
                        direct_rouse,
                        conflict_wait,
                        conflict_cb,
                    ],
                )
            )
            return executor

        return make_args

    def on_full_match(
        self,
        target: str | list[str],
        logic_mode: LogicMode = LogicMode.OR,
        checker: Optional["BotChecker"] = None,
        priority: PriorLevel = PriorLevel.MEAN,
        block: bool = False,
        temp: bool = False,
        session_rule: Optional["SessionRule"] = None,
        session_hold: bool = False,
        direct_rouse: bool = False,
        conflict_wait: bool = False,
        conflict_cb: Optional[Callable[[], Coroutine[Any, Any, None]]] = None,
    ):
        """注册一个字符串全匹配的消息事件处理器

        `target` 为字符串时，只进行一次全匹配，即判断是否匹配成功。
        `target` 为字符串列表时，所有字符串都进行全匹配，再将所有结果使用给定
        `logic_mode` 计算是否匹配成功。

        消息必须匹配成功才能被进一步处理。

        :param target: 匹配目标
        :param logic_mode: `target` 为 `list[str]` 时的计算模式
        :param checker: 使用的检查器，为空则默认通过检查
        :param priority: 优先级
        :param block: 是否进行优先级阻断
        :param temp: 是否是一次性的
        :param session_rule: 会话规则，为空则不使用会话规则
        :param session_hold: 处理方法结束后是否保留会话（有会话规则才可启用）
        :param direct_rouse: 会话暂停时，是否允许不检查就唤醒会话（有会话规则才可启用）
        :param conflict_wait: 会话冲突时，是否需要事件等待处理（有会话规则才可启用）
        :param conflict_cb: 会话冲突时，运行的回调（有会话规则才可启用，`conflict_wait=True`，此参数无效）
        """
        full_matcher = FullMatcher(target, logic_mode)

        def make_args(
            executor: Callable[[], Coroutine[Any, Any, None]]
        ) -> Callable[[], Coroutine[Any, Any, None]]:
            self.__handler_args__.append(
                EventHandlerArgs(
                    executor=executor,
                    type=MsgEventHandler,
                    params=[
                        full_matcher,
                        None,
                        checker,
                        priority,
                        block,
                        temp,
                        session_rule,
                        session_hold,
                        direct_rouse,
                        conflict_wait,
                        conflict_cb,
                    ],
                )
            )
            return executor

        return make_args

    def on_end_match(
        self,
        target: str | list[str],
        logic_mode: LogicMode = LogicMode.OR,
        checker: Optional["BotChecker"] = None,
        priority: PriorLevel = PriorLevel.MEAN,
        block: bool = False,
        temp: bool = False,
        session_rule: Optional["SessionRule"] = None,
        session_hold: bool = False,
        direct_rouse: bool = False,
        conflict_wait: bool = False,
        conflict_cb: Optional[Callable[[], Coroutine[Any, Any, None]]] = None,
    ):
        """注册一个字符串结尾匹配的消息事件处理器

        `target` 为字符串时，只进行一次结尾匹配，即判断是否匹配成功。
        `target` 为字符串列表时，所有字符串都进行结尾匹配，再将所有结果使用给定
        `logic_mode` 计算是否匹配成功。

        消息必须匹配成功才能被进一步处理。

        :param target: 匹配目标
        :param logic_mode: `target` 为 `list[str]` 时的计算模式
        :param checker: 使用的检查器，为空则默认通过检查
        :param priority: 优先级
        :param block: 是否进行优先级阻断
        :param temp: 是否是一次性的
        :param session_rule: 会话规则，为空则不使用会话规则
        :param session_hold: 处理方法结束后是否保留会话（有会话规则才可启用）
        :param direct_rouse: 会话暂停时，是否允许不检查就唤醒会话（有会话规则才可启用）
        :param conflict_wait: 会话冲突时，是否需要事件等待处理（有会话规则才可启用）
        :param conflict_cb: 会话冲突时，运行的回调（有会话规则才可启用，`conflict_wait=True`，此参数无效）
        """
        end_matcher = EndMatcher(target, logic_mode)

        def make_args(
            executor: Callable[[], Coroutine[Any, Any, None]]
        ) -> Callable[[], Coroutine[Any, Any, None]]:
            self.__handler_args__.append(
                EventHandlerArgs(
                    executor=executor,
                    type=MsgEventHandler,
                    params=[
                        end_matcher,
                        None,
                        checker,
                        priority,
                        block,
                        temp,
                        session_rule,
                        session_hold,
                        direct_rouse,
                        conflict_wait,
                        conflict_cb,
                    ],
                )
            )
            return executor

        return make_args

    def on_regex_match(
        self,
        target: str,
        checker: Optional["BotChecker"] = None,
        priority: PriorLevel = PriorLevel.MEAN,
        block: bool = False,
        temp: bool = False,
        session_rule: Optional["SessionRule"] = None,
        session_hold: bool = False,
        direct_rouse: bool = False,
        conflict_wait: bool = False,
        conflict_cb: Optional[Callable[[], Coroutine[Any, Any, None]]] = None,
    ):
        """注册一个字符串正则匹配的消息事件处理器

        消息必须匹配成功才能被进一步处理。

        :param target: 匹配目标的正则表达式，在匹配时，它应该可以使 :meth:`re.findall` 不返回空列表
        :param checker: 使用的检查器，为空则默认通过检查
        :param priority: 优先级
        :param block: 是否进行优先级阻断
        :param temp: 是否是一次性的
        :param session_rule: 会话规则，为空则不使用会话规则
        :param session_hold: 处理方法结束后是否保留会话（有会话规则才可启用）
        :param direct_rouse: 会话暂停时，是否允许不检查就唤醒会话（有会话规则才可启用）
        :param conflict_wait: 会话冲突时，是否需要事件等待处理（有会话规则才可启用）
        :param conflict_cb: 会话冲突时，运行的回调（有会话规则才可启用，`conflict_wait=True`，此参数无效）
        """
        regex_matcher = RegexMatcher(target)

        def make_args(
            executor: Callable[[], Coroutine[Any, Any, None]]
        ) -> Callable[[], Coroutine[Any, Any, None]]:
            self.__handler_args__.append(
                EventHandlerArgs(
                    executor=executor,
                    type=MsgEventHandler,
                    params=[
                        regex_matcher,
                        None,
                        checker,
                        priority,
                        block,
                        temp,
                        session_rule,
                        session_hold,
                        direct_rouse,
                        conflict_wait,
                        conflict_cb,
                    ],
                )
            )
            return executor

        return make_args

    def on_request(
        self,
        checker: Optional["BotChecker"] = None,
        priority: PriorLevel = PriorLevel.MEAN,
        block: bool = False,
        temp: bool = False,
        session_rule: Optional["SessionRule"] = None,
        session_hold: bool = False,
        direct_rouse: bool = False,
        conflict_wait: bool = False,
        conflict_cb: Optional[Callable[[], Coroutine[Any, Any, None]]] = None,
    ):
        """注册一个请求事件处理器

        :param checker: 使用的检查器，为空则默认通过检查
        :param priority: 优先级
        :param block: 是否进行优先级阻断
        :param temp: 是否是一次性的
        :param session_rule: 会话规则，为空则不使用会话规则
        :param session_hold: 处理方法结束后是否保留会话（有会话规则才可启用）
        :param direct_rouse: 会话暂停时，是否允许不检查就唤醒会话（有会话规则才可启用）
        :param conflict_wait: 会话冲突时，是否需要事件等待处理（有会话规则才可启用）
        :param conflict_cb: 会话冲突时，运行的回调（有会话规则才可启用，`conflict_wait=True`，此参数无效）
        """

        def make_args(
            executor: Callable[[], Coroutine[Any, Any, None]]
        ) -> Callable[[], Coroutine[Any, Any, None]]:
            self.__handler_args__.append(
                EventHandlerArgs(
                    executor=executor,
                    type=ReqEventHandler,
                    params=[
                        checker,
                        priority,
                        block,
                        temp,
                        session_rule,
                        session_hold,
                        direct_rouse,
                        conflict_wait,
                        conflict_cb,
                    ],
                )
            )
            return executor

        return make_args

    def on_friend_request(
        self,
        checker: Optional["BotChecker"] = None,
        priority: PriorLevel = PriorLevel.MEAN,
        block: bool = False,
        temp: bool = False,
        session_rule: Optional["SessionRule"] = None,
        session_hold: bool = False,
        direct_rouse: bool = False,
        conflict_wait: bool = False,
        conflict_cb: Optional[Callable[[], Coroutine[Any, Any, None]]] = None,
    ):
        """注册一个好友请求事件处理器

        :param checker: 使用的检查器，为空则默认通过检查
        :param priority: 优先级
        :param block: 是否进行优先级阻断
        :param temp: 是否是一次性的
        :param session_rule: 会话规则，为空则不使用会话规则
        :param session_hold: 处理方法结束后是否保留会话（有会话规则才可启用）
        :param direct_rouse: 会话暂停时，是否允许不检查就唤醒会话（有会话规则才可启用）
        :param conflict_wait: 会话冲突时，是否需要事件等待处理（有会话规则才可启用）
        :param conflict_cb: 会话冲突时，运行的回调（有会话规则才可启用，`conflict_wait=True`，此参数无效）
        """
        friend_checker = FriendReqChecker()
        wrapped_checker: FriendReqChecker | "WrappedChecker"
        if checker is not None:
            wrapped_checker = friend_checker & checker
        else:
            wrapped_checker = friend_checker

        def make_args(
            executor: Callable[[], Coroutine[Any, Any, None]]
        ) -> Callable[[], Coroutine[Any, Any, None]]:
            self.__handler_args__.append(
                EventHandlerArgs(
                    executor=executor,
                    type=ReqEventHandler,
                    params=[
                        wrapped_checker,
                        priority,
                        block,
                        temp,
                        session_rule,
                        session_hold,
                        direct_rouse,
                        conflict_wait,
                        conflict_cb,
                    ],
                )
            )
            return executor

        return make_args

    def on_group_request(
        self,
        checker: Optional["BotChecker"] = None,
        priority: PriorLevel = PriorLevel.MEAN,
        block: bool = False,
        temp: bool = False,
        session_rule: Optional["SessionRule"] = None,
        session_hold: bool = False,
        direct_rouse: bool = False,
        conflict_wait: bool = False,
        conflict_cb: Optional[Callable[[], Coroutine[Any, Any, None]]] = None,
    ):
        """注册一个加群请求事件处理器

        :param checker: 使用的检查器，为空则默认通过检查
        :param priority: 优先级
        :param block: 是否进行优先级阻断
        :param temp: 是否是一次性的
        :param session_rule: 会话规则，为空则不使用会话规则
        :param session_hold: 处理方法结束后是否保留会话（有会话规则才可启用）
        :param direct_rouse: 会话暂停时，是否允许不检查就唤醒会话（有会话规则才可启用）
        :param conflict_wait: 会话冲突时，是否需要事件等待处理（有会话规则才可启用）
        :param conflict_cb: 会话冲突时，运行的回调（有会话规则才可启用，`conflict_wait=True`，此参数无效）
        """
        group_checker = GroupReqChecker()
        wrapped_checker: GroupReqChecker | "WrappedChecker"
        if checker is not None:
            wrapped_checker = group_checker & checker
        else:
            wrapped_checker = group_checker

        def make_args(
            executor: Callable[[], Coroutine[Any, Any, None]]
        ) -> Callable[[], Coroutine[Any, Any, None]]:
            self.__handler_args__.append(
                EventHandlerArgs(
                    executor=executor,
                    type=ReqEventHandler,
                    params=[
                        wrapped_checker,
                        priority,
                        block,
                        temp,
                        session_rule,
                        session_hold,
                        direct_rouse,
                        conflict_wait,
                        conflict_cb,
                    ],
                )
            )
            return executor

        return make_args

    def on_notice(
        self,
        type: Literal[
            "group_upload",
            "group_admin",
            "group_decrease",
            "group_increase",
            "group_ban",
            "friend_add",
            "group_recall",
            "friend_recall",
            "group_card",
            "offline_file",
            "client_status",
            "essence",
            "notify",
            "honor",
            "poke",
            "lucky_king",
            "title",
            "ALL",
        ] = "ALL",
        checker: Optional["BotChecker"] = None,
        priority: PriorLevel = PriorLevel.MEAN,
        block: bool = False,
        temp: bool = False,
        session_rule: Optional["SessionRule"] = None,
        session_hold: bool = False,
        direct_rouse: bool = False,
        conflict_wait: bool = False,
        conflict_cb: Optional[Callable[[], Coroutine[Any, Any, None]]] = None,
    ):
        """注册一个通知事件处理器

        :param type: 通知的类型，为 "ALL" 时接受所有通知
        :param checker: 使用的检查器，为空则默认通过检查
        :param priority: 优先级
        :param block: 是否进行优先级阻断
        :param temp: 是否是一次性的
        :param session_rule: 会话规则，为空则不使用会话规则
        :param session_hold: 处理方法结束后是否保留会话（有会话规则才可启用）
        :param direct_rouse: 会话暂停时，是否允许不检查就唤醒会话（有会话规则才可启用）
        :param conflict_wait: 会话冲突时，是否需要事件等待处理（有会话规则才可启用）
        :param conflict_cb: 会话冲突时，运行的回调（有会话规则才可启用，`conflict_wait=True`，此参数无效）
        """
        type_checker = NoticeTypeChecker(type)
        wrapped_checker: NoticeTypeChecker | "WrappedChecker"
        if checker is not None:
            wrapped_checker = type_checker & checker
        else:
            wrapped_checker = type_checker

        def make_args(
            executor: Callable[[], Coroutine[Any, Any, None]]
        ) -> Callable[[], Coroutine[Any, Any, None]]:
            self.__handler_args__.append(
                EventHandlerArgs(
                    executor=executor,
                    type=NoticeEventHandler,
                    params=[
                        wrapped_checker,
                        priority,
                        block,
                        temp,
                        session_rule,
                        session_hold,
                        direct_rouse,
                        conflict_wait,
                        conflict_cb,
                    ],
                )
            )
            return executor

        return make_args

    def on_meta_event(
        self,
        checker: Optional["BotChecker"] = None,
        priority: PriorLevel = PriorLevel.MEAN,
        block: bool = False,
        temp: bool = False,
        session_rule: Optional["SessionRule"] = None,
        session_hold: bool = False,
        direct_rouse: bool = False,
        conflict_wait: bool = False,
        conflict_cb: Optional[Callable[[], Coroutine[Any, Any, None]]] = None,
    ):
        """注册一个元事件处理器

        :param checker: 使用的检查器，为空则默认通过检查
        :param priority: 优先级
        :param block: 是否进行优先级阻断
        :param temp: 是否是一次性的
        :param session_rule: 会话规则，为空则不使用会话规则
        :param session_hold: 处理方法结束后是否保留会话（有会话规则才可启用）
        :param direct_rouse: 会话暂停时，是否允许不检查就唤醒会话（有会话规则才可启用）
        :param conflict_wait: 会话冲突时，是否需要事件等待处理（有会话规则才可启用）
        :param conflict_cb: 会话冲突时，运行的回调（有会话规则才可启用，`conflict_wait=True`，此参数无效）
        """

        def make_args(
            executor: Callable[[], Coroutine[Any, Any, None]]
        ) -> Callable[[], Coroutine[Any, Any, None]]:
            self.__handler_args__.append(
                EventHandlerArgs(
                    executor=executor,
                    type=MetaEventHandler,
                    params=[
                        checker,
                        priority,
                        block,
                        temp,
                        session_rule,
                        session_hold,
                        direct_rouse,
                        conflict_wait,
                        conflict_cb,
                    ],
                )
            )
            return executor

        return make_args

    def on_signal(self, namespace: str, signal: str):
        """注册一个信号处理方法

        本方法作为异步函数的装饰器使用，此时可注册一个函数为信号处理方法。

        .. code:: python

           # 假设存在一个名为 plugin 的变量，是 BotPlugin 实例
           # 为在命名空间 BaseUtils 中，名为 txt2img 的信号绑定处理方法：
           @plugin.on_signal("BaseUtils", "txt2img")
           async def get_img_of_txt(text: str, format: Any) -> bytes:
               # melobot 对被装饰函数的参数类型和返回值没有限制
               # 接下来是具体逻辑
               ...
           # 在这个示例中，具体的功能是为其他插件提供转换大段文本为图片的能力，因为大段文本不便于发送

        .. admonition:: 注意
           :class: caution

           在一个 bot 实例的范围内，同命名空间同名称的信号，只能注册一个处理方法。

        :param namespace: 信号的命名空间
        :param signal: 信号的名称
        """

        def make_args(
            func: Callable[P, Coroutine[Any, Any, Any]]
        ) -> Callable[P, Coroutine[Any, Any, Any]]:
            self.__signal_args__.append(
                PluginSignalHandlerArgs(func, namespace, signal)
            )
            return func

        return make_args

    def on_share(
        self, namespace: str, id: str, reflector: Optional[Callable[[], Any]] = None
    ):
        """注册一个共享对象，同时绑定它的值获取方法

        本方法可作为异步函数的装饰器使用，此时被装饰函数就是共享对象的值获取方法：

        .. code:: python

           # 假设存在一个名为 plugin 的变量，是 BotPlugin 实例
           # 在命名空间 HelpUtils 中，注册一个名为 all_helps 的共享对象，且绑定值获取方法：
           @plugin.on_share("HelpUtils", "all_helps")
           async def get_all_helps() -> str:
               # melobot 对被装饰函数的要求：无参数，但必须有返回值
               return ALL_HELPS_INFO_STR
           # 在这个示例中，具体的功能是在插件间共享 “所有插件的帮助文本” 这一数据

        当然，值获取方法较为简单时，直接传参即可：

        .. code:: python

           # 最后一个参数不能给定具体的值，必须为一个同步函数
           plugin.on_share("HelpUtils", "all_helps", lambda: ALL_HELPS_INFO_STR)

        .. admonition:: 注意
           :class: caution

           在一个 bot 实例的范围内，同命名空间同名称的共享对象，只能注册一个。

        :param namespace: 共享对象的命名空间
        :param id: 共享对象的 id 标识
        :param reflector: 为空时，本方法当作异步函数的装饰器使用；否则应该直接使用，此处提供共享对象值获取的反射函数
        """
        if reflector is not None:
            self.__share_args__.append(ShareObjArgs(namespace, id, to_async(reflector)))
            return

        def make_args(
            func: Callable[[], Coroutine[Any, Any, Any]]
        ) -> Callable[[], Coroutine[Any, Any, Any]]:
            self.__share_args__.append(ShareObjArgs(namespace, id, func))
            return func

        return make_args

    def on_share_affected(self, namespace: str, id: str):
        """为一个共享对象注册回调方法

        本方法作为异步函数的装饰器使用，此时可为一个共享对象注册回调方法。

        .. code:: python

           # 假设存在一个名为 plugin 的变量，是 BotPlugin 实例
           # 为在命名空间 HelpUtils 中，名为 all_helps 的共享对象绑定回调方法：
           @plugin.on_share_affected("HelpUtils", "all_helps")
           async def add_a_help(text: str) -> bool:
               # melobot 对被装饰函数的参数类型和返回值没有限制
               # 接下来是具体逻辑
               ...
           # 此回调用于被其他插件触发，为它们提供“影响”共享对象的能力，
           # 在这个示例中，具体的功能是让其他插件可以添加一条自己的帮助信息，但是有所校验

        .. admonition:: 注意
           :class: caution

           在一个 bot 实例的范围内，同命名空间同名称的共享对象，只能注册一个回调方法。
           而且这个共享对象必须在本插件通过 :meth:`on_share` 注册（共享对象注册、共享对象回调注册先后顺序不重要）

        :param namespace: 共享对象的命名空间
        :param id: 共享对象的 id 标识
        """

        def make_args(
            func: Callable[P, Coroutine[Any, Any, Any]]
        ) -> Callable[P, Coroutine[Any, Any, Any]]:
            self.__share_cb_args__.append(ShareObjCbArgs(namespace, id, func))
            return func

        return make_args

    def on_bot_life(self, type: BotLife):
        """注册 bot 在某个生命周期的 hook 方法

        本方法作为异步函数的装饰器使用，此时可注册一个函数为 bot 生命周期 hook 方法。

        .. code:: python

           # 假设存在一个名为 plugin 的变量，是 BotPlugin 实例
           # 我们希望这个插件，在 bot 连接器建立连接后给某人发一条消息
           @plugin.on_bot_life(BotLife.CONNECTED)
           async def say_hi() -> None:
               # melobot 对被装饰函数的要求：无参数，返回空值
               await send_custom_msg("Hello~", isPrivate=True, userId=xxxxx)
           # 在这个示例中，bot 登录上号后，便会向 xxxxx 发送一条 Hello~ 消息

        :param type: bot 生命周期类型枚举值
        """

        def make_args(
            func: Callable[P, Coroutine[Any, Any, None]]
        ) -> Callable[P, Coroutine[Any, Any, None]]:
            self.__hook_args__.append(BotHookRunnerArgs(func, type))
            return func

        return make_args

    @property
    def on_plugins_loaded(self):
        """注册 bot 在 :attr:`.BotLife.LOADED` 生命周期的 hook 方法

        本方法作为异步函数的装饰器使用。用法与 :class:`on_bot_life` 类似。
        """
        return self.on_bot_life(BotLife.LOADED)

    @property
    def on_connected(self):
        """注册 bot 在 :attr:`.BotLife.CONNECTED` 生命周期的 hook 方法

        本方法作为异步函数的装饰器使用。用法与 :class:`on_bot_life` 类似。
        """
        return self.on_bot_life(BotLife.CONNECTED)

    @property
    def on_before_close(self):
        """注册 bot 在 :attr:`.BotLife.BEFORE_CLOSE` 生命周期的 hook 方法

        本方法作为异步函数的装饰器使用。用法与 :class:`on_bot_life` 类似。
        """
        return self.on_bot_life(BotLife.BEFORE_CLOSE)

    @property
    def on_before_stop(self):
        """注册 bot 在 :attr:`.BotLife.BEFORE_STOP` 生命周期的 hook 方法

        本方法作为异步函数的装饰器使用。用法与 :class:`on_bot_life` 类似。
        """
        return self.on_bot_life(BotLife.BEFORE_STOP)

    @property
    def on_event_built(self):
        """注册 bot 在 :attr:`.BotLife.EVENT_BUILT` 生命周期的 hook 方法

        本方法作为异步函数的装饰器使用。用法与 :class:`on_bot_life` 类似。
        """
        return self.on_bot_life(BotLife.EVENT_BUILT)

    @property
    def on_action_presend(self):
        """注册 bot 在 :attr:`.BotLife.ACTION_PRESEND` 生命周期的 hook 方法

        本方法作为异步函数的装饰器使用。用法与 :class:`on_bot_life` 类似。
        """
        return self.on_bot_life(BotLife.ACTION_PRESEND)



import asyncio

from ..base.abc import BotLife
from ..base.exceptions import BotHookError, get_better_exc
from ..base.tools import get_rich_str, to_task
from ..base.typing import TYPE_CHECKING, Any, Callable, Coroutine

if TYPE_CHECKING:
    from ..utils.logger import Logger


class HookRunner:
    """bot hook 运行器"""

    def __init__(
        self, type: BotLife, func: Callable[..., Coroutine[Any, Any, None]]
    ) -> None:
        self.cb: Callable[..., Coroutine[Any, Any, None]] = func
        self.type: BotLife = type


class BotHookBus:
    """bot hook 总线"""

    def __init__(self) -> None:
        self.store: dict[BotLife, list[HookRunner]] = {
            v: [] for k, v in BotLife.__members__.items()
        }
        self.logger: "Logger"

    def _bind(self, logger: "Logger") -> None:
        self.logger = logger

    def register(
        self, hook_type: BotLife, hook_func: Callable[..., Coroutine[Any, Any, None]]
    ) -> None:
        if hook_type not in self.store.keys():
            raise BotHookError(
                f"尝试添加一个生命周期 hook 方法，但是其指定的类型 {hook_type} 不存在"
            )
        runner = HookRunner(hook_type, hook_func)
        self.store[hook_type].append(runner)

    async def _run_on_ctx(self, runner: HookRunner, *args: Any, **kwargs: Any) -> None:
        try:
            await runner.cb(*args, **kwargs)
        except Exception as e:
            func_name = runner.cb.__qualname__
            self.logger.error(f"hook 方法 {func_name} 发生异常")
            self.logger.error("异常回溯栈：\n" + get_better_exc(e))
            self.logger.error("异常点局部变量：\n" + get_rich_str(locals()))

    async def emit(
        self, hook_type: BotLife, *args: Any, wait: bool = False, **kwargs: Any
    ) -> None:
        if not wait:
            for runner in self.store[hook_type]:
                to_task(self._run_on_ctx(runner, *args, **kwargs))
        else:
            tasks = []
            for runner in self.store[hook_type]:
                tasks.append(to_task(self._run_on_ctx(runner, *args, **kwargs)))
            if len(tasks):
                await asyncio.wait(tasks)



import asyncio
import os
import sys
from contextvars import ContextVar, Token
from copy import deepcopy

from ..base.abc import AbstractConnector
from ..base.exceptions import BotRuntimeError, DuplicateError, get_better_exc
from ..base.tools import get_rich_str, to_task
from ..base.typing import (
    TYPE_CHECKING,
    Any,
    BotLife,
    Coroutine,
    Literal,
    Optional,
    Void,
)
from ..context.session import SESSION_LOCAL, BotSessionManager
from ..controller.dispatcher import BotDispatcher
from ..controller.responder import BotResponder
from ..meta import MELOBOT_LOGO, MELOBOT_LOGO_LEN, MetaInfo
from ..models.event import BotEventBuilder
from ..plugin.handler import EVENT_HANDLER_MAP
from ..plugin.init import BotPlugin, PluginLoader, PluginProxy
from ..plugin.ipc import PluginBus, PluginStore
from ..utils.logger import BotLogger, Logger, NullLogger
from .hook import BotHookBus

if TYPE_CHECKING:
    from ..plugin.ipc import ShareObject

if sys.platform != "win32":
    import uvloop

    asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())

_ = BotLogger("__MELOBOT_ENTRY__", no_tag=True)
__ = MetaInfo()
for line in MELOBOT_LOGO.split("\n"):
    _.info(line)
_.info(f" 运行版本：{__.VER} | 平台：{__.PLATFORM}")
_.info("-" * MELOBOT_LOGO_LEN)


def _safe_run_bot(main_coro: Coroutine[Any, Any, None]) -> None:
    try:
        asyncio.set_event_loop(asyncio.get_event_loop())
        asyncio.run(main_coro)
    except KeyboardInterrupt:
        pass


class MeloBot:
    """bot 类。该类实例化即一个 bot 实例"""

    BOTS: dict[str, "MeloBot"] = {}

    def __init__(self, name: str) -> None:
        """初始化一个 bot 实例

        :param name: bot 实例的名字（唯一）
        """
        if name in MeloBot.BOTS.keys():
            raise DuplicateError(f"命名为 {name} 的 bot 实例已存在，请改名避免冲突")
        MeloBot.BOTS[name] = self

        #: bot 的名字（唯一）
        self.name: str = name
        #: 元信息
        self.info: MetaInfo = MetaInfo()
        #: bot 的日志器
        self.logger: Logger
        #: bot 的连接器
        self.connector: AbstractConnector

        self.__plugins: dict[str, BotPlugin] = {}
        self._loader = PluginLoader
        self._ctx_manager = BotSessionManager
        self._event_builder = BotEventBuilder
        self._life: list[asyncio.Task]
        self._plugin_store = PluginStore()
        self._plugin_bus = PluginBus()
        self._bot_bus = BotHookBus()
        self._responder = BotResponder()
        self._dispatcher = BotDispatcher()

        self.__init_flag__: bool = False
        self.__run_flag__: bool = False

    def init(
        self,
        connector: AbstractConnector,
        enable_log: bool = True,
        logger_name: Optional[str] = None,
        log_level: Literal["DEBUG", "ERROR", "INFO", "WARNING", "CRITICAL"] = "INFO",
        log_to_console: bool = True,
        log_to_dir: Optional[str] = None,
        custom_logger: Optional[Logger] = None,
    ) -> "MeloBot":
        """初始化 bot 实例

        :param connector: 使 bot 实例工作的连接器
        :param enable_log: 是否启用日志
        :param logger_name: 日志器名称, 为空则使用 bot 实例名字
        :param log_level: 日志器日志等级
        :param log_to_console: 日志是否输出到控制台
        :param log_to_dir: 保存日志文件的目录，为空则不保存
        :param custom_logger: 自定义日志器对象。若不为空将使用该日志器，并忽略其他所有日志相关参数
        :return: bot 实例（因此支持链式调用）
        """
        if connector._ref_flag:
            raise DuplicateError("bot 初始化时，不可使用已被其他 bot 实例使用的连接器")
        self.connector = connector
        self.connector._ref_flag = True
        if custom_logger is not None:
            self.logger = custom_logger
        elif not enable_log:
            self.logger = NullLogger(f"__MELOBOT_NULL_{self.name}__")
        else:
            self.logger = BotLogger(
                self.name if logger_name is None else logger_name,
                log_level,
                log_to_console,
                log_to_dir,
            )
        self.logger.debug(f"连接器已初始化，类型：{connector.__class__.__name__}")

        self._dispatcher._bind(
            EVENT_HANDLER_MAP, self._bot_bus, self._ctx_manager, self.logger
        )
        self._plugin_bus._bind(self.logger)
        self._bot_bus._bind(self.logger)
        self._responder._bind(self.logger, self.connector)
        self.connector._bind(
            self._dispatcher,
            self._responder,
            self._event_builder,
            self._bot_bus,
            self.logger,
        )
        self.logger.debug("bot 初始化完成，各核心组件已初始化")
        self.__init_flag__ = True
        return self

    def load_plugin(self, plugin_target: str | BotPlugin) -> "MeloBot":
        """为 bot 实例加载插件

        :param plugin_target: 插件目标，可传入插件对象或插件包（一个插件即是一个 python package）的路径
        :return: bot 实例（因此支持链式调用）
        """
        if not self.__init_flag__:
            raise BotRuntimeError("bot 尚未初始化，不能加载插件")

        plugin_dir = None if not isinstance(plugin_target, str) else plugin_target
        if plugin_dir is None:
            self.logger.debug(f"尝试加载插件 {plugin_target}")
        else:
            self.logger.debug(f"尝试加载来自 {plugin_dir} 的插件")
        plugin = self._loader.load(plugin_target)
        exist_plugin = self.__plugins.get(plugin.__id__)
        if exist_plugin is not None:
            self.logger.error(
                f"加载插件出错：插件 id 重复, 重复的 id 为：{plugin.__id__}，已取消该插件加载"
            )
            return self

        handlers = []
        for _ in plugin.__handler_args__:
            handler = _.type(_.executor, plugin, self.logger, *_.params)
            self._ctx_manager.register(handler)
            handlers.append(handler)
        for _ in plugin.__share_args__:
            self._plugin_store.create_so(_.reflector, _.namespace, _.id)
        for _ in plugin.__share_cb_args__:
            self._plugin_store.bind_cb(_.namespace, _.id, _.cb)
        for _ in plugin.__signal_args__:
            self._plugin_bus.register(_.namespace, _.signal, _.func)
        for _ in plugin.__hook_args__:
            self._bot_bus.register(_.type, _.func)

        self.__plugins[plugin.__id__] = plugin
        self._dispatcher.add_handlers(handlers)
        self.logger.info(f"成功加载插件：{plugin.__id__}")
        return self

    def load_plugins(self, plugins_dir: str) -> "MeloBot":
        """为 bot 实例批量加载插件

        :param plugins_dir: 传入包含所有插件包（一个插件即是一个 python package）的目录
        :return: bot 实例（因此支持链式调用）
        """
        if not self.__init_flag__:
            raise BotRuntimeError("bot 尚未初始化，不能加载插件")

        self.logger.debug(f"尝试从目录 {plugins_dir} 批量加载插件")
        items = os.listdir(plugins_dir)
        for item in items:
            path = os.path.join(plugins_dir, item)
            if os.path.isdir(path) and os.path.basename(path) != "__pycache__":
                self.load_plugin(path)
        return self

    async def _run(self) -> None:
        if not self.__init_flag__:
            raise BotRuntimeError("bot 尚未初始化，不能启动")
        if self.__run_flag__:
            raise BotRuntimeError("bot 已在运行，无需再次启动")
        if len(self.__plugins) == 0:
            self.logger.warning("没有加载任何插件，bot 将不会有任何操作")

        bot_token = BOT_LOCAL._add_ctx(self)
        await self._bot_bus.emit(BotLife.LOADED)
        self.logger.debug("LOADED hook 已完成")
        try:
            async with self.connector:
                self._dispatcher._set_ready()
                self._responder._set_ready()
                self.connector._set_ready()
                self._life = await self.connector._alive_tasks()
                self.__run_flag__ = True
                self.logger.info("bot 开始正常运行")
                self.logger.debug(
                    f"使用的连接器类型：{self.connector.__class__.__name__}"
                )
                await asyncio.wait(self._life)
        except Exception as e:
            self.logger.error(f"bot 核心无法继续运行。异常：{e}")
            self.logger.error("异常回溯栈：\n" + get_better_exc(e))
            self.logger.error("异常点局部变量：\n" + get_rich_str(locals()))
        finally:
            await self._bot_bus.emit(BotLife.BEFORE_STOP, wait=True)
            self.logger.debug("BEFORE_STOP hook 已完成")
            self.logger.info("bot 已清理运行时资源")
            BOT_LOCAL._del_ctx(bot_token)
            self.__run_flag__ = False

    def run(self) -> None:
        """运行 bot 实例"""
        _safe_run_bot(self._run())

    async def close(self) -> None:
        """停止 bot 实例"""
        if not self.__run_flag__:
            raise BotRuntimeError("bot 尚未运行，无需停止")

        await self._bot_bus.emit(BotLife.BEFORE_CLOSE, wait=True)
        self.logger.debug("BEFORE_CLOSE hook 已完成")
        for task in self._life:
            task.cancel()

    def is_activate(self) -> bool:
        """判断 bot 实例是否在非 slack 状态

        slack 状态启用后仅会禁用行为操作的发送，无其他影响

        :return: 是否在非 slack 状态
        """
        if not self.__init_flag__:
            raise BotRuntimeError("bot 尚未初始化，不能执行此方法")

        return not self.connector.slack

    def activate(self) -> "MeloBot":
        """使 bot 实例退出 slack 状态

        slack 状态启用后仅会禁用行为操作的发送，无其他影响

        :return: bot 实例（因此支持链式调用）
        """
        if not self.__init_flag__:
            raise BotRuntimeError("bot 尚未初始化，不能执行此方法")

        self.connector.slack = False
        self.logger.debug("bot 已进入 activated 状态")
        return self

    def slack(self) -> "MeloBot":
        """使 bot 实例进入 slack 状态

        slack 状态启用后仅会禁用行为操作的发送，无其他影响

        :return: bot 实例（因此支持链式调用）
        """
        if not self.__init_flag__:
            raise BotRuntimeError("bot 尚未初始化，不能执行此方法")

        self.connector.slack = True
        self.logger.debug("bot 已进入 slack 状态")
        return self

    def get_plugins(self) -> dict[str, PluginProxy]:
        """获得 bot 实例所有插件的信息

        :return: 所有插件的信息
        """
        if not self.__init_flag__:
            raise BotRuntimeError("bot 尚未初始化，不能执行此方法")

        return {name: deepcopy(p.__proxy__) for name, p in self.__plugins.items()}

    def emit_signal(
        self, namespace: str, signal: str, *args: Any, wait: bool = False, **kwargs: Any
    ) -> Any:
        """在本 bot 实例范围内发起一个信号

        :param namespace: 触发信号的命名空间
        :param signal: 触发信号的名字
        :param wait: 是否等待信号处理方法处理完毕
        :param args: 传递给信号处理方法的 args
        :param kwargs: 传递给信号处理方法的 kwargs
        :return: 不等待信号处理方法处理，只返回 :obj:`None`；若等待则返回运行结果
        """
        if not self.__init_flag__:
            raise BotRuntimeError("bot 尚未初始化，不能执行此方法")

        self.logger.debug(f"bot 信号总线信号触发：{namespace}.{signal} | wait: {wait}")
        self.logger.debug("本次信号触发参数：\n" + get_rich_str(args))
        return self._plugin_bus.emit(namespace, signal, *args, wait=wait, **kwargs)

    def get_share(self, namespace: str, id: str) -> "ShareObject":
        """获得 bot 实例内的共享对象

        :param namespace: 共享对象的命名空间
        :param id: 共享对象的标识 id
        :return: 获得到的共享对象
        """
        if not self.__init_flag__:
            raise BotRuntimeError("bot 尚未初始化，不能执行此方法")

        self.logger.debug(f"获取共享对象行为：{namespace}.{id}")
        return self._plugin_store.get(namespace, id)

    @classmethod
    def start(cls, *bots: "MeloBot") -> None:
        """同时运行多个 bot 实例

        :param bots: 要运行的 bot 实例
        """

        async def bots_run():
            tasks = []
            for bot in bots:
                tasks.append(to_task(bot._run()))
            try:
                await asyncio.wait(tasks)
            except asyncio.CancelledError:
                pass

        _safe_run_bot(bots_run())

    @classmethod
    async def unicast(
        cls,
        target: str,
        namespace: str,
        signal: str,
        *args: Any,
        wait: bool = False,
        **kwargs: Any,
    ) -> None:
        """在指定的 bot 实例范围发起信号，即单播

        :param target: 单播的目标 bot 实例的名字
        :param namespace: 信号的命名空间
        :param signal: 信号的名字
        :param wait: 是否等待信号处理方法处理完毕
        :param args: 传递给信号处理方法的 args
        :param kwargs: 传递给信号处理方法的 kwargs
        """
        bot = cls.BOTS.get(target)
        if bot is None:
            raise BotRuntimeError(f"单播指定的 bot 实例 {target} 不存在")

        b_token = BOT_LOCAL._add_ctx(bot)
        s_token = SESSION_LOCAL._add_ctx(Void)
        try:
            await bot.emit_signal(namespace, signal, *args, **kwargs, wait=wait)
        except AttributeError as e:
            if "Void" in e.__str__():
                raise BotRuntimeError(
                    "多播或单播时，bot 和 session 上下文的传递将会被阻隔，如需要请将它们作为参数显式传递"
                )
            else:
                raise e
        finally:
            BOT_LOCAL._del_ctx(b_token)
            SESSION_LOCAL._del_ctx(s_token)

    @classmethod
    async def multicast(
        cls,
        targets: list[str] | Literal["ALL"],
        namespace: str,
        signal: str,
        *args: Any,
        self_exclude: bool = True,
        wait: bool = False,
        **kwargs: Any,
    ) -> None:
        """在指定的多个 bot 实例范围发起信号，即多播

        :param targets: 多个 bot 实例的名字列表，为 "ALL" 时代表向所有 bot 多播，即广播
        :param namespace: 信号的命名空间
        :param signal: 信号的名字
        :param self_exclude: 是否在多播时排除自己
        :param wait: 是否等待信号处理方法处理完毕
        :param args: 传递给信号处理方法的 args
        :param kwargs: 传递给信号处理方法的 kwargs
        """
        if isinstance(targets, list):
            _targets = targets
        else:
            _targets = list(cls.BOTS.keys())
            if self_exclude:
                _targets.remove(BOT_LOCAL.name)
        tasks = []
        for name in _targets:
            tasks.append(
                to_task(
                    cls.unicast(name, namespace, signal, *args, **kwargs, wait=wait)
                )
            )
        if len(tasks):
            await asyncio.wait(tasks)


class BotLocal:
    """bot 实例自动上下文"""

    __slots__ = tuple(
        list(filter(lambda x: not (len(x) >= 2 and x[:2] == "__"), dir(MeloBot)))
        + ["__storage__"]
    )

    def __init__(self) -> None:
        object.__setattr__(self, "__storage__", ContextVar("melobot_ctx"))
        self.__storage__: ContextVar["MeloBot"]

    def __setattr__(self, __name: str, __value: Any) -> None:
        setattr(self.__storage__.get(), __name, __value)

    def __getattr__(self, __name: str) -> Any:
        return getattr(self.__storage__.get(), __name)

    def _add_ctx(self, ctx: "MeloBot") -> Token:
        return self.__storage__.set(ctx)

    def _del_ctx(self, token: Token) -> None:
        self.__storage__.reset(token)


BOT_LOCAL = BotLocal()
BotSessionManager._bind(BOT_LOCAL)



from ..base.abc import BotChecker
from ..base.exceptions import BotCheckerError
from ..base.typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Coroutine,
    Literal,
    Optional,
    User,
)

if TYPE_CHECKING:
    from ..models.event import MessageEvent, NoticeEvent, RequestEvent


class MsgLvlChecker(BotChecker):
    """消息分级检查器

    主要分 主人、超级用户、白名单用户、普通用户、黑名单用户 五级
    """

    def __init__(
        self,
        level: User,
        owner: Optional[int] = None,
        super_users: Optional[list[int]] = None,
        white_users: Optional[list[int]] = None,
        black_users: Optional[list[int]] = None,
        ok_cb: Optional[Callable[[], Coroutine[Any, Any, None]]] = None,
        fail_cb: Optional[Callable[[], Coroutine[Any, Any, None]]] = None,
    ) -> None:
        """初始化一个消息分级检查器

        :param level: 允许的等级（>= 此等级才能通过校验）
        :param owner: 主人的 qq 号
        :param super_users: 超级用户 qq 号列表
        :param white_users: 白名单用户 qq 号列表
        :param black_users: 黑名单用户 qq 号列表
        :param ok_cb: 检查通过的回调
        :param fail_cb: 检查不通过的回调
        """
        super().__init__(ok_cb, fail_cb)
        self.check_lvl = level

        self.owner = owner
        self.su_list = super_users if super_users is not None else []
        self.white_list = white_users if white_users is not None else []
        self.black_list = black_users if black_users is not None else []

    def _get_level(self, event: "MessageEvent") -> User:
        """获得事件对应的登记."""
        qid = event.sender.id

        if qid in self.black_list:
            return User.BLACK
        elif qid == self.owner:
            return User.OWNER
        elif qid in self.su_list:
            return User.SU
        elif qid in self.white_list:
            return User.WHITE
        else:
            return User.USER

    async def check(self, event: "MessageEvent") -> bool:  # type: ignore
        if not event.is_msg_event():
            return False
        e_level = self._get_level(event)
        status = 0 < e_level.value and e_level.value >= self.check_lvl.value
        if status and self.ok_cb is not None:
            await self.ok_cb()
        if not status and self.fail_cb is not None:
            await self.fail_cb()
        return status


class GroupMsgLvlChecker(MsgLvlChecker):
    """群聊消息分级检查器

    基本功能与 :class:`MsgLvlChecker` 一致。但增加了白名单机制，
    不提供 `white_groups` 参数默认拒绝所有群聊消息。

    对所有私聊消息校验不通过。
    """

    def __init__(
        self,
        level: User,
        owner: Optional[int] = None,
        super_users: Optional[list[int]] = None,
        white_users: Optional[list[int]] = None,
        black_users: Optional[list[int]] = None,
        white_groups: Optional[list[int]] = None,
        ok_cb: Optional[Callable[[], Coroutine[Any, Any, None]]] = None,
        fail_cb: Optional[Callable[[], Coroutine[Any, Any, None]]] = None,
    ) -> None:
        """初始化一个群聊消息分级检查器

        :param level: 允许的等级（>= 此等级才能通过校验）
        :param owner: 主人的 qq 号
        :param super_users: 超级用户 qq 号列表
        :param white_users: 白名单用户 qq 号列表
        :param black_users: 黑名单用户 qq 号列表
        :param white_groups: 白名单群号列表（不在其中的群不通过校验）
        :param ok_cb: 检查通过的回调
        :param fail_cb: 检查不通过的回调
        """
        super().__init__(
            level, owner, super_users, white_users, black_users, ok_cb, fail_cb
        )
        self.white_group_list = white_groups if white_groups is not None else []

    async def check(self, event: "MessageEvent") -> bool:  # type: ignore
        if not event.is_msg_event():
            return False
        if (
            not event.is_group()
            or len(self.white_group_list) == 0
            or (event.group_id not in self.white_group_list)
        ):
            return False
        return await super().check(event)


class PrivateMsgLvlChecker(MsgLvlChecker):
    """私聊消息分级检查器

    基本功能与 :class:`MsgLvlChecker` 一致。

    对所有群聊消息校验不通过。
    """

    def __init__(
        self,
        level: User,
        owner: Optional[int] = None,
        super_users: Optional[list[int]] = None,
        white_users: Optional[list[int]] = None,
        black_users: Optional[list[int]] = None,
        ok_cb: Optional[Callable[[], Coroutine[Any, Any, None]]] = None,
        fail_cb: Optional[Callable[[], Coroutine[Any, Any, None]]] = None,
    ) -> None:
        """初始化一个私聊消息分级检查器

        :param level: 允许的等级（>= 此等级才能通过校验）
        :param owner: 主人的 qq 号
        :param super_users: 超级用户 qq 号列表
        :param white_users: 白名单用户 qq 号列表
        :param black_users: 黑名单用户 qq 号列表
        :param ok_cb: 检查通过的回调
        :param fail_cb: 检查不通过的回调
        """
        super().__init__(
            level, owner, super_users, white_users, black_users, ok_cb, fail_cb
        )

    async def check(self, event: "MessageEvent") -> bool:  # type: ignore
        if not event.is_msg_event():
            return False
        if not event.is_private():
            return False
        return await super().check(event)


class MsgCheckerGen:
    """消息分级检查器的生成器

    预先存储检查依据（各等级数据），指定检查等级后，可返回一个 :class:`MsgLvlChecker` 类的对象
    """

    def __init__(
        self,
        owner: Optional[int] = None,
        super_users: Optional[list[int]] = None,
        white_users: Optional[list[int]] = None,
        black_users: Optional[list[int]] = None,
        white_groups: Optional[list[int]] = None,
        ok_cb: Optional[Callable[[], Coroutine[Any, Any, None]]] = None,
        fail_cb: Optional[Callable[[], Coroutine[Any, Any, None]]] = None,
    ) -> None:
        """初始化一个消息分级检查器的生成器

        :param owner: 主人的 qq 号
        :param super_users: 超级用户 qq 号列表
        :param white_users: 白名单用户 qq 号列表
        :param black_users: 黑名单用户 qq 号列表
        :param white_groups: 白名单群号列表（不在其中的群不通过校验）
        :param ok_cb: 检查通过的回调（这将自动附加到生成的检查器上）
        :param fail_cb: 检查不通过的回调（这将自动附加到生成的检查器上）
        """
        self.owner = owner
        self.su_list = super_users if super_users is not None else []
        self.white_list = white_users if white_users is not None else []
        self.black_list = black_users if black_users is not None else []
        self.white_group_list = white_groups if white_groups is not None else []

        self.united_ok_cb = ok_cb
        self.united_fail_cb = fail_cb

    def gen_base(self, level: User = User.USER) -> MsgLvlChecker:
        """根据内部依据和给定等级，生成一个 :class:`MsgLvlChecker` 对象

        :param level: 允许的等级（>= 此等级才能通过校验）
        :return: 消息分级检查器
        """
        return MsgLvlChecker(
            level,
            self.owner,
            self.su_list,
            self.white_list,
            self.black_list,
            self.united_ok_cb,
            self.united_fail_cb,
        )

    def gen_group(self, level: User = User.USER) -> GroupMsgLvlChecker:
        """根据内部依据和给定等级，生成一个 :class:`GroupMsgLvlChecker` 对象

        :param level: 允许的等级（>= 此等级才能通过校验）
        :return: 群聊消息分级检查器
        """
        return GroupMsgLvlChecker(
            level,
            self.owner,
            self.su_list,
            self.white_list,
            self.black_list,
            self.white_group_list,
            self.united_ok_cb,
            self.united_fail_cb,
        )

    def gen_private(self, level: User = User.USER) -> PrivateMsgLvlChecker:
        """根据内部依据和给定等级，生成一个 :class:`PrivateMsgLvlChecker` 对象

        :param level: 允许的等级（>= 此等级才能通过校验）
        :return: 私聊消息分级检查器
        """
        return PrivateMsgLvlChecker(
            level,
            self.owner,
            self.su_list,
            self.white_list,
            self.black_list,
            self.united_ok_cb,
            self.united_fail_cb,
        )


class AtChecker(BotChecker):
    """艾特消息检查器"""

    def __init__(self, qid: Optional[int] = None) -> None:
        """初始化一个艾特消息检查器

        :param qid: 被艾特的 qq 号。为空则接受所有艾特消息；不为空则只接受指定 qid 被艾特的艾特消息
        """
        super().__init__(None, None)
        self.qid = qid if qid is not None else None

    async def check(self, event: "MessageEvent") -> bool:  # type: ignore
        if not event.is_msg_event():
            return False
        id_list = event.get_cq_params("at", "qq")
        if self.qid is None:
            status = len(id_list) > 0
        else:
            for id in id_list:
                if id == self.qid:
                    status = True
                    break
            status = False
        return status


class FriendReqChecker(BotChecker):
    """好友请求检查器"""

    def __init__(self) -> None:
        """初始化一个好友请求检查器

        只有是请求类型中的好友请求才会通过检查
        """
        super().__init__(None, None)

    async def check(self, event: "RequestEvent") -> bool:  # type: ignore
        if not event.is_req_event():
            return False
        status = event.is_friend_req()
        return status


class GroupReqChecker(BotChecker):
    """加群请求检查器"""

    def __init__(self) -> None:
        """初始化一个加群请求检查器

        只有是请求类型中的加群请求才会通过检查
        """
        super().__init__(None, None)

    async def check(self, event: "RequestEvent") -> bool:  # type: ignore
        if not event.is_req_event():
            return False
        status = event.is_group_req()
        return status


class NoticeTypeChecker(BotChecker):
    """通知检查器"""

    SUB_TYPES = [
        "group_upload",
        "group_admin",
        "group_decrease",
        "group_increase",
        "group_ban",
        "friend_add",
        "group_recall",
        "friend_recall",
        "group_card",
        "offline_file",
        "client_status",
        "essence",
        "notify",
        "honor",
        "poke",
        "lucky_king",
        "title",
        "ALL",
    ]

    def __init__(
        self,
        sub_type: Literal[
            "group_upload",
            "group_admin",
            "group_decrease",
            "group_increase",
            "group_ban",
            "friend_add",
            "group_recall",
            "friend_recall",
            "group_card",
            "offline_file",
            "client_status",
            "essence",
            "notify",
            "honor",
            "poke",
            "lucky_king",
            "title",
            "ALL",
        ],
    ) -> None:
        """初始化一个通知检查器

        只有是给定类型的通知事件才会通过检查

        :param sub_type: 通知的类型
        """
        super().__init__(None, None)
        if sub_type not in NoticeTypeChecker.SUB_TYPES:
            raise BotCheckerError(f"通知事件类型校验器的子类型 {sub_type} 不合法")
        self.sub_type = sub_type

    async def check(self, event: "NoticeEvent") -> bool:  # type: ignore
        if not event.is_notice_event():
            return False
        if self.sub_type == "ALL":
            status = True
        else:
            check_method = getattr(event, "is_" + self.sub_type)
            status = check_method()
        return status



import sys

from better_exceptions import ExceptionFormatter


class BotException(Exception):
    """Bot 异常基类

    在使用 melobot 编写代码时，若需要抛出异常，可以有意抛出该类。
    表明这是你设计的异常情况，而不是完全预期之外的异常。
    """

    def __init__(self, msg: str):
        super().__init__(self, msg)
        self.err = msg

    def __str__(self):
        return self.err


class DuplicateError(BotException):
    def __init__(self, msg: str):
        super().__init__(msg)


class BotRuntimeError(BotException):
    def __init__(self, msg: str):
        super().__init__(msg)


class PluginInitError(BotException):
    def __init__(self, msg: str):
        super().__init__(msg)


class PluginBuildError(BotException):
    def __init__(self, msg: str):
        super().__init__(msg)


class BotConnectFailed(BotException):
    def __init__(self, msg: str):
        super().__init__(msg)


class BotHookError(BotException):
    def __init__(self, msg: str):
        super().__init__(msg)


class EventHandlerError(BotException):
    def __init__(self, msg: str):
        super().__init__(msg)


class BotCheckerError(BotException):
    def __init__(self, msg: str):
        super().__init__(msg)


class BotMatcherError(BotException):
    def __init__(self, msg: str):
        super().__init__(msg)


class TryFlagFailed(BotException):
    def __init__(self, msg: str):
        super().__init__(msg)


class ShareObjError(BotException):
    def __init__(self, msg: str):
        super().__init__(msg)


class PluginSignalError(BotException):
    def __init__(self, msg: str):
        super().__init__(msg)


class BotActionError(BotException):
    def __init__(self, msg: str):
        super().__init__(msg)


class BotSessionError(BotException):
    def __init__(self, msg: str):
        super().__init__(msg)


class BotBaseUtilsError(BotException):
    def __init__(self, msg: str):
        super().__init__(msg)


class ArgParseError(BotException):
    def __init__(self, msg: str):
        super().__init__(msg)


class ArgFormatInitError(BotException):
    def __init__(self, msg: str):
        super().__init__(msg)


class ArgVerifyFailed(BotException):
    def __init__(self, msg: str = ""):
        super().__init__(msg)


class ArgLackError(BotException):
    def __init__(self, msg: str = ""):
        super().__init__(msg)


class DirectRetSignal(BotException):
    def __init__(self, msg: str = ""):
        super().__init__(msg)


class SessionHupTimeout(BotException):
    """会话暂停的超时异常"""

    def __init__(self, msg: str = ""):
        super().__init__(msg)


_EXC_FORMATTER = ExceptionFormatter(colored=False)


def get_better_exc(e: Exception) -> str:
    """返回生成更好的异常字符串."""
    return "".join(
        _EXC_FORMATTER.format_exception(e.__class__, e, sys.exc_info()[2])
    ).strip("\n")



import base64
import io
import json
import os
from typing import Any, Dict, List, Tuple

import requests
from PIL import Image
from requests import Response


class SdApi:

    class ApiBaseException(Exception):
        def __init__(self, err: str):
            super().__init__()
            self.err = f"[SdApiException] {err}"
            self.origin_err = err

        def __str__(self):
            return self.err

    class ValueError(ApiBaseException):
        def __init__(self, err: str):
            super().__init__(err)

    class ResponseError(ApiBaseException):
        def __init__(self, err: str):
            super().__init__(err)

    class ApiStore:
        def __init__(self) -> None:
            self.models: List[str]
            self.loras: List[Tuple[str, str]]
            self.samplers: List[str]
            self.upsclaers: List[str]
            self.options: Dict[str, str]
            self.cmd_args: Dict[str, str]
            self.scripts: Dict[str, List[str]]
            self.vaes: List[str]

        def fill(
            self,
            models,
            loras,
            samplers,
            upscalers,
            options,
            cmd_args,
            scripts,
            vae_list,
        ) -> None:
            (
                self.models,
                self.loras,
                self.samplers,
                self.upsclaers,
                self.options,
                self.cmd_args,
                self.scripts,
                self.vaes,
            ) = (
                models,
                loras,
                samplers,
                upscalers,
                options,
                cmd_args,
                scripts,
                vae_list,
            )

    def __init__(self) -> None:
        self.config_path = (
            os.sep.join(os.path.join(__file__).split(os.sep)[:-1])
            + os.sep
            + "config.json"
        )

        with open(self.config_path, encoding="utf-8") as fp:
            config = json.load(fp)
            self.endpoint = config["endpoint"]
            self.auth_user = config["username"]
            self.auth_pwd = config["password"]
            self.negative_prompt_default = config["negative_defaults"]
        self.session = requests.Session()
        self.session.auth = (self.auth_user, self.auth_pwd)

        self.store = SdApi.ApiStore()
        self.refresh_store(self.store)

    @property
    def progress(self) -> float:
        return self._get("/progress")["progress"]

    @property
    def job_count(self) -> int:
        return self._get("/progress")["state"]["job_count"]

    def _get(
        self,
        path: str,
        params: dict = None,
        headers: dict = None,
        pureResp: bool = False,
    ) -> dict | list | Response:
        resp = self.session.get(self.endpoint + path, params=params, headers=headers)
        if resp.status_code != 200:
            raise SdApi.ResponseError(
                f"error code: {resp.status_code}, content: {resp.json()}"
            )
        return resp if pureResp else resp.json()

    def _post(
        self,
        path: str,
        data: dict = None,
        headers: dict = None,
        pureResp: bool = False,
        jsonSendMode: bool = True,
    ) -> dict | list | Response:
        if jsonSendMode:
            resp = self.session.post(self.endpoint + path, json=data, headers=headers)
        else:
            resp = self.session.post(self.endpoint + path, data=data, headers=headers)

        if resp.status_code != 200:
            raise SdApi.ResponseError(
                f"error code: {resp.status_code}, content: {resp.json()}"
            )
        return resp if pureResp else resp.json()

    def _get_models(self) -> List[str]:
        content = self._get("/sd-models")
        return [model_info["title"] for model_info in content]

    def _get_loras(self) -> List[Tuple[str, str]]:
        content = self._get("/loras")
        return [(lora_info["name"], lora_info["alias"]) for lora_info in content]

    def _get_samplers(self) -> List[str]:
        content = self._get("/samplers")
        return [sampler_info["name"] for sampler_info in content]

    def _get_upsclaers(self) -> List[str]:
        content = self._get("/upscalers")
        return [upscaler_info["name"] for upscaler_info in content]

    def _get_options(self) -> Dict[str, str]:
        return self._get("/options")

    def _get_cmd_args(self) -> Dict[str, str]:
        return self._get("/cmd-flags")

    def _get_scripts(self) -> Dict[str, List[str]]:
        return self._get("/scripts")

    def _get_vaes(self) -> Dict[str, List[str]]:
        vae_infos = self._get("/sd-vae")
        return [info["model_name"] for info in vae_infos]

    def refresh_store(self, store: ApiStore) -> None:
        store.fill(
            self._get_models(),
            self._get_loras(),
            self._get_samplers(),
            self._get_upsclaers(),
            self._get_options(),
            self._get_cmd_args(),
            self._get_scripts(),
            self._get_vaes(),
        )
        # print("已刷新本地缓存")

    def skip(self) -> None:
        self._post("/skip")

    def interrupt(self) -> None:
        self._post("/interrupt")

    def modify_option(self, args: Tuple[str, str] | Dict[str, Any]) -> None:
        if isinstance(args, tuple):
            args = {args[0]: args[1]}
        self._post("/options", data=args)

    def txt2img(
        self,
        prompt: str,
        negative_prompt: str = None,
        width: int = 512,
        height: int = 512,
        steps: int = 20,
        sampler_name: str = "DPM++ 3M SDE Karras",
        cfg_scale: float = 7,
        seed: int = None,
        subseed: int = None,
        subseed_strength: float = None,
        enable_hr: bool = None,
        denoising_strength: float = None,
        hr_scale: int = None,
        hr_upscaler: str = None,
        hr_second_pass_steps: int = None,
        script_name: str = None,
        script_args: list = None,
        override_settings: Dict[str, str] = None,
        limit_access: bool = True,
    ) -> Tuple[List[Image.Image], dict, str]:
        """
        返回 Image 列表，生成参数字典，生成信息
        """
        if limit_access:
            if len(prompt) > 1000:
                raise SdApi.ValueError("正向提示词限制长度 1000")
            if negative_prompt is not None and len(negative_prompt) > 500:
                raise SdApi.ValueError("负向提示词限制长度 500")
            if width < 100 or width > 960 or height < 100 or height > 960:
                raise SdApi.ValueError("宽高限制： [100, 960]")
            if width * height > 518400:
                raise SdApi.ValueError("宽高乘积不得大于 518400")
            if steps < 10 or steps > 60:
                raise SdApi.ValueError("步数限制：[10, 60]")
            if cfg_scale < 1 or cfg_scale > 30:
                raise SdApi.ValueError("cfg限制：[1, 30]")
            if subseed_strength is not None and (
                subseed_strength < 0 or subseed_strength > 1
            ):
                raise SdApi.ValueError("差异随机种子限制：[0, 1]")
            if enable_hr is not None or override_settings is not None:
                raise SdApi.ValueError("无权使用高清修复或临时配置功能")
            (
                enable_hr,
                denoising_strength,
                hr_scale,
                hr_upscaler,
                hr_second_pass_steps,
                override_settings,
            ) = (None, None, None, None, None, None)
            script_name, script_args = "censorscript", [True, True]

        if negative_prompt is None:
            negative_prompt = self.negative_prompt_default

        copyed_args = locals().copy()
        copyed_args.pop("self")
        copyed_args.pop("limit_access")
        copyed_args = {k: v for k, v in copyed_args.items() if v is not None}

        content = self._post("/txt2img", data=copyed_args)
        return (
            [
                Image.open(io.BytesIO(base64.b64decode(image.split(",", 1)[0])))
                for image in content["images"]
            ],
            content["parameters"],
            content["info"],
        )



from .action import *
from .session import (
    AttrSessionRule,
    any_event,
    dispose,
    get_store,
    meta_event,
    msg_args,
    msg_event,
    msg_text,
    notice_event,
    pause,
    req_event,
)

__all__ = (
    "send_custom_msg",
    "send",
    "send_custom_forward",
    "send_forward",
    "msg_recall",
    "get_msg",
    "get_forward_msg",
    "get_image",
    "send_like",
    "group_kick",
    "group_ban",
    "group_whole_ban",
    "set_group_admin",
    "set_group_card",
    "set_group_name",
    "group_leave",
    "set_group_title",
    "set_friend_add",
    "set_group_add",
    "get_login_info",
    "get_stranger_info",
    "get_friend_list",
    "get_group_info",
    "get_group_list",
    "get_group_member_info",
    "get_group_member_list",
    "get_group_honor",
    "check_send_image",
    "check_send_record",
    "get_onebot_version",
    "get_onebot_status",
    "take_custom_action",
    "make_action",
    "send_wait",
    "send_reply",
    "finish",
    "reply_finish",
    "AttrSessionRule",
    "any_event",
    "get_store",
    "meta_event",
    "msg_args",
    "msg_event",
    "msg_text",
    "notice_event",
    "pause",
    "req_event",
    "dispose",
)



import json
import re
import time

from ..base.abc import BotEvent
from ..base.typing import Any, CQMsgDict, Literal, Optional, Type
from .cq import _get_cq, _get_cq_params, to_cq_arr, to_cq_str


class BotEventBuilder:
    @staticmethod
    def build(rawEvent: dict | str) -> BotEvent:
        if isinstance(rawEvent, str):
            raw: dict[str, str | float | int] = json.loads(rawEvent)
        else:
            raw = rawEvent

        etype = raw.get("post_type")
        if etype in ("message_sent", "message"):
            return MessageEvent(raw)
        elif etype == "request":
            return RequestEvent(raw)
        elif etype == "notice":
            return NoticeEvent(raw)
        elif etype == "meta_event":
            return MetaEvent(raw)
        else:
            return ResponseEvent(raw)


class MessageEvent(BotEvent):
    """消息事件类型

    .. admonition:: 提示
       :class: tip

       一般无需手动实例化该类，多数情况会直接使用本类对象，或将本类用作类型注解。
    """

    def __init__(self, rawEvent: dict) -> None:
        super().__init__(rawEvent)
        #: 收到事件的机器人 qq 号
        self.bot_id: int = rawEvent.get("self_id")  # type: ignore
        #: 消息事件的消息 id
        self.id: int
        #: 消息事件的发送者数据结构
        self.sender: MessageEvent.Sender
        #: 消息事件的来源群号，私聊时此属性也存在并为 :obj:`None`
        self.group_id: Optional[int]
        #: 使用 CQ 字符串表示的，本事件的所有消息内容
        self.raw_content: str
        #: 使用消息段对象列表表示的，本事件的所有消息内容
        self.content: list[CQMsgDict]
        #: 本事件的所有消息内容中，纯文本消息的合并字符串
        self.text: str
        #: 消息字体
        self.font: int
        #: 临时消息的来源标记，不是临时消息时为 :obj:`None`
        self.temp_src: Optional[
            Literal[
                "群聊",
                "QQ咨询",
                "查找",
                "QQ电影",
                "热聊",
                "验证消息",
                "多人聊天",
                "约会",
                "通讯录",
            ]
        ]

        self._init()

    @property
    def time(self) -> int:
        """事件发生时刻的时间戳"""
        return self.raw.get("time")  # type: ignore

    @property
    def type(self) -> Literal["message"]:
        """事件的类型（返回指定字面量）"""
        return "message"

    def _init(self) -> None:
        rawEvent = self.raw

        self._cq_regex = re.compile(r"\[CQ:.*?\]")
        self.id = rawEvent["message_id"]
        self.raw_content = self._format_to_str(rawEvent["raw_message"])
        self.content = self._format_to_array(rawEvent["message"])
        self.text = self._get_text(self.content)
        self.font = rawEvent["font"]

        self.temp_src = None
        temp_src = rawEvent.get("temp_source")
        if temp_src:
            self.temp_src = MessageEvent._TEMP_SRC_MAP[temp_src]  # type: ignore

        self.sender = MessageEvent.Sender(
            rawEvent=rawEvent,
            isGroup=self.is_group(),
            isGroupAnonym=self.is_group_anonym(),
        )

        self.group_id = None
        if self.is_group():
            self.group_id = rawEvent["group_id"]

    def _format_to_str(self, content: list | str) -> str:
        """对外部零信任，强制转换为 cq 字符串格式."""
        if not isinstance(content, str):
            return to_cq_str(content)
        else:
            return content

    def _format_to_array(self, content: list | str) -> list[CQMsgDict]:
        """对外部零信任，强制转换为消息段格式."""
        if not isinstance(content, str):
            for item in content:
                if item["type"] == "text":
                    continue
                for k, v in item["data"].items():
                    if not isinstance(v, str):
                        continue
                    if v.isdigit() or (len(v) >= 2 and v[0] == "-" and v[1:].isdigit()):
                        item["data"][k] = int(v)
                        continue
                    try:
                        item["data"][k] = float(v)
                    except Exception:
                        pass
            return content
        else:
            return to_cq_arr(content)

    def _get_text(self, content: list[CQMsgDict]) -> str:
        """获取消息中所有文本消息，返回合并字符串."""
        text_list: list[str] = []
        for item in content:
            if item["type"] == "text":
                text_list.append(item["data"]["text"])  # type: ignore
        return "".join(text_list)

    def get_cq(self, cq_type: str) -> list[CQMsgDict]:
        """提取指定类型的消息组成消息段对象列表

        :param cq_type: 消息段对象类型
        :return: 消息段对象列表
        """
        return _get_cq(self.content, cq_type)

    def get_cq_params(
        self, cq_type: str, param: str, type: Optional[Type[Any]] = None
    ) -> list[Any]:
        """提取指定类型的消息中的指定参数

        当没有任何对应类型的消息时，为空列表。如果有对应类型的消息，但是指定的参数不存在，
        则在列表中产生 :obj:`None` 值.

        可以指定 type 来强制转换类型，不填则使用智能解析出的类型

        使用示例如下：

        .. code:: python

           # 假设 event 是 MessageEvent 对象，即象征一个消息事件
           params = event.get_cq_params('at', 'qq')
           # params 将是此事件中，所有 at 类型消息的 "qq" 参数所组成的列表

        :param cq_type: 消息类型
        :param param: 该类消息中的参数名
        :param type: 类型转换使用的类
        :return: 参数值列表
        """
        return _get_cq_params(self.content, cq_type, param, type)

    def is_private(self) -> bool:
        """是否为私聊消息（注意群临时会话属于该类别）"""
        return self.raw["message_type"] == "private"

    def is_friend(self) -> bool:
        """是否为好友消息"""
        return (
            self.raw["message_type"] == "private" and self.raw["sub_type"] == "friend"
        )

    def is_group(self) -> bool:
        """是否为群消息（正常群消息、群匿名消息、群自身消息、群系统消息属于该类型）"""
        return self.raw["message_type"] == "group"

    def is_group_normal(self) -> bool:
        """是否为正常群消息"""
        return self.raw["message_type"] == "group" and self.raw["sub_type"] == "normal"

    def is_group_anonym(self) -> bool:
        """是否为匿名群消息"""
        return (
            self.raw["message_type"] == "group" and self.raw["sub_type"] == "anonymous"
        )

    def is_group_self(self) -> bool:
        """是否为群自身消息（即 bot 自己群中发的消息）"""
        return (
            self.raw["message_type"] == "group" and self.raw["sub_type"] == "group_self"
        )

    def is_group_temp(self) -> bool:
        """是否为群临时会话（属于私聊的一种）"""
        return self.raw["message_type"] == "private" and self.raw["sub_type"] == "group"

    def is_temp(self) -> bool:
        """是否为临时会话（属于私聊的一种）"""
        return "temp_source" in self.raw.keys()

    def is_group_notice(self) -> bool:
        """是否为群中的\"系统消息\" """
        return self.raw["message_type"] == "group" and self.raw["sub_type"] == "notice"

    _TEMP_SRC_MAP = {
        0: "群聊",
        1: "QQ咨询",
        2: "查找",
        3: "QQ电影",
        4: "热聊",
        6: "验证消息",
        7: "多人聊天",
        8: "约会",
        9: "通讯录",
    }

    class Sender:
        """消息事件发送者数据结构

        .. admonition:: 提示
           :class: tip

           一般无需手动实例化该类，多数情况会直接使用本类对象，或将本类用作类型注解。
        """

        def __init__(self, rawEvent: dict, isGroup: bool, isGroupAnonym: bool) -> None:
            self._rawEvent = rawEvent
            self._isGroup = isGroup
            self._isGroupAnonym = isGroupAnonym

            #: 发送者的 qq 号
            self.id: int
            #: 发送者昵称
            self.nickname: Optional[str]
            #: 发送者性别
            self.sex: Optional[str]
            #: 发送者年龄
            self.age: Optional[int]
            #: 发送者的群名片
            self.group_card: str
            #: 发送者的群中角色
            self.group_role: Literal["owner", "admin", "member", "anonymous"]
            #: 发送者的群头衔
            self.group_title: str
            #: 发送者的地区
            self.group_area: str
            #: 发送者的群等级
            self.group_level: str

            #: 匿名发送者的 id（此属性只在事件为群匿名消息事件时存在）
            self.anonym_id: int
            #: 匿名发送者的名字（此属性只在事件为群匿名消息事件时存在）
            self.anonym_name: str
            #: 匿名发送者的匿名 flag（此属性只在事件为群匿名消息事件时存在）
            self.anonym_flag: str

            # 对应私聊可能缺失 sender 的情况
            if len(rawEvent["sender"]) == 0:
                self._gen_empty()
                return

            self.id = rawEvent["sender"]["user_id"]
            self.nickname = rawEvent["sender"]["nickname"]
            self.sex = (
                rawEvent["sender"]["sex"]
                if "sex" in rawEvent["sender"].keys()
                else None
            )
            self.age = (
                rawEvent["sender"]["age"]
                if "age" in rawEvent["sender"].keys()
                else None
            )
            if isGroup:
                if isGroupAnonym:
                    self.group_card = ""
                    self.group_area = rawEvent["sender"]["area"]
                    self.group_level = rawEvent["sender"]["level"]
                    self.group_role = "anonymous"
                    self.group_title = ""
                    self.anonym_id = rawEvent["anonymous"]["id"]
                    self.anonym_name = rawEvent["anonymous"]["name"]
                    self.anonym_flag = rawEvent["anonymous"]["flag"]
                else:
                    self.group_card = rawEvent["sender"]["card"]
                    self.group_area = (
                        rawEvent["sender"]["area"]
                        if "area" in rawEvent["sender"].keys()
                        else ""
                    )
                    self.group_level = rawEvent["sender"]["level"]
                    self.group_role = rawEvent["sender"]["role"]
                    self.group_title = rawEvent["sender"]["title"]

        def _gen_empty(self) -> None:
            """当传入的 rawEvent 缺失 sender 时，把属性直接置空."""
            self.id = self._rawEvent["user_id"]
            self.nickname = None
            self.sex = None
            self.age = None

        def is_group_owner(self) -> bool:
            """判断是否为群主，若不是或不是群类型消息，返回 False"""
            if not self._isGroup:
                return False
            return self.group_role == "owner"

        def is_group_admin(self) -> bool:
            """判断是否为群管理（包含群主），若不是或不是群类型消息，返回 False"""
            if not self._isGroup:
                return False
            return self.group_role == "admin" or self.group_role == "owner"

        def only_group_member(self) -> bool:
            """判断是否只是群员（注意只是群员，不包括群主、管理和匿名），若不是或不是群类型消息，返回 False"""
            if not self._isGroup:
                return False
            return self.group_role == "member"

        def is_anonym_member(self) -> bool:
            """判断是否是群匿名，若不是或不是群类型消息，返回 False"""
            if not self._isGroup:
                return False
            return self.group_role == "anonymous"

        def is_bot(self) -> bool:
            """判断消息是否是bot自己发送的"""
            return self.id == self._rawEvent["self_id"]


class RequestEvent(BotEvent):
    """请求事件类型，对应两种可能：加好友请求和加群请求

    .. admonition:: 提示
       :class: tip

       一般无需手动实例化该类，多数情况会直接使用本类对象，或将本类用作类型注解。
    """

    def __init__(self, rawEvent: dict) -> None:
        super().__init__(rawEvent)

        #: 收到事件的机器人 qq 号
        self.bot_id: int = rawEvent.get("self_id")  # type: ignore
        #: 事件的来源 qq 号
        self.from_id: int
        #: 事件的来源群号，请求来源于私聊时此属性为 :obj:`None`
        self.from_group_id: Optional[int]
        #: 加群或加好友的验证消息
        self.req_comment: str
        #: 加群或加好友请求的 flag，调用相关 API 时，需要使用
        self.req_flag: str
        #: 加群请求类型：直接加入和邀请，请求来源于私聊时此属性为 :obj:`None`
        self.group_req_type: Optional[Literal["add", "invite"]]

        self._init()

    @property
    def time(self) -> int:
        """事件发生时刻的时间戳"""
        return self.raw.get("time")  # type: ignore

    @property
    def type(self) -> Literal["request"]:
        """事件的类型（返回指定字面量）"""
        return "request"

    def _init(self) -> None:
        rawEvent = self.raw
        self.group_req_type = None
        self.from_group_id = None

        if self.is_friend_req():
            self.from_id = rawEvent["user_id"]
            self.req_comment = rawEvent["comment"]
            self.req_flag = rawEvent["flag"]
        elif self.is_group_req():
            self.group_req_type = rawEvent["sub_type"]
            self.from_id = rawEvent["user_id"]
            self.from_group_id = rawEvent["group_id"]
            self.req_comment = rawEvent["comment"]
            self.req_flag = rawEvent["flag"]

    def is_friend_req(self) -> bool:
        """是否为加好友请求"""
        return self.raw["request_type"] == "friend"

    def is_group_req(self) -> bool:
        """是否为加群请求"""
        return self.raw["request_type"] == "group"


class NoticeEvent(BotEvent):
    """通知事件类型

    .. admonition:: 提示
       :class: tip

       一般无需手动实例化该类，多数情况会直接使用本类对象，或将本类用作类型注解。

    .. admonition:: 注意
       :class: caution

       受 onebot 协议实现项目的影响，以及对通知事件本身复杂性的考虑，通知事件的所有实例属性，
       都只会在特定类别的通知事件中存在。想要使用这些属性，最好先经过实际验证。

       例如你想要处理一个群禁言通知事件，建议先绑定“群禁言通知事件”的事件处理器，
       然后调试查看事件都有哪些属性。这样就可以确定对“群禁言”这一类别的通知事件哪些属性可用。
    """

    def __init__(self, rawEvent: dict) -> None:
        super().__init__(rawEvent)
        #: 收到事件的机器人 qq 号
        self.bot_id: int = rawEvent.get("self_id")  # type: ignore
        # 修复某些 onebot 协议实现，user_id 缺失的问题
        if "target_id" in rawEvent.keys() and "user_id" not in rawEvent:
            rawEvent["user_id"] = rawEvent["target_id"]

        #: 通知作用者或主体方的 id，如被禁言的一方
        self.notice_user_id: int
        #: 通知若发生在群中的群 id
        self.notice_group_id: int
        #: 通知发起者或操作方的 id，如禁言别人的管理员
        self.notice_operator_id: int
        #: 通知涉及消息时的消息 id
        self.notice_msg_id: int
        #: 入群通知事件的入群类型：管理员同意或管理员邀请
        self.join_group_type: Literal["approve", "invite"]
        #: 退群通知事件的退群类型：自退群、踢出、bot 账号被踢
        self.leave_group_type: Literal["leave", "kick", "kick_me"]
        #: 群管理员变动通知事件类型：设置管理员或取消管理员
        self.admin_change_type: Literal["set", "unset"]
        #: 文件上传通知事件的文件信息数据结构
        self.file: NoticeEvent.File
        #: 群禁言通知事件的类型：禁言、解除禁言
        self.group_ban_type: Literal["ban", "lift_ban"]
        #: 群禁言通知事件的禁言时长
        self.ban_time: int
        #: 群荣誉变更通知事件的荣誉类型：龙王、群聊之火、快乐源泉
        self.honor_change_type: Literal["talkactive", "performer", "emotion"]
        #: 群头衔变更通知事件的新头衔
        self.new_title: str
        #: 群名片更新通知事件的旧名片，名片为空时对应属性为空字符串
        self.old_card: str
        #: 群名片更新通知事件的新名片，名片为空时对应属性为空字符串
        self.new_card: str
        #: 客户端在线状态变更的通知事件的，客户端信息数据结构
        self.client: NoticeEvent.Client
        #: 精华消息变更通知事件的类型：添加或删除
        self.essence_change_type: Literal["add", "delete"]

        self._init()

    @property
    def time(self) -> int:
        return self.raw.get("time")  # type: ignore

    @property
    def type(self) -> Literal["notice"]:
        return "notice"

    def _init(self) -> None:
        """外部确认为该类型事件时，调用此方法。"""
        rawEvent = self.raw

        if self.is_friend_recall():
            self.notice_user_id = rawEvent["user_id"]
            self.notice_msg_id = rawEvent["message_id"]
        elif self.is_group_recall():
            self.notice_user_id = rawEvent["user_id"]
            self.notice_group_id = rawEvent["group_id"]
            self.notice_operator_id = rawEvent["operator_id"]
            self.notice_msg_id = rawEvent["message_id"]
        elif self.is_group_increase():
            self.join_group_type = rawEvent["sub_type"]
            self.notice_user_id = rawEvent["user_id"]
            self.notice_group_id = rawEvent["group_id"]
            self.notice_operator_id = rawEvent["operator_id"]
        elif self.is_group_decrease():
            self.leave_group_type = rawEvent["sub_type"]
            self.notice_user_id = rawEvent["user_id"]
            self.notice_group_id = rawEvent["group_id"]
            self.notice_operator_id = rawEvent["operator_id"]
        elif self.is_group_admin():
            self.admin_change_type = rawEvent["sub_type"]
            self.notice_user_id = rawEvent["user_id"]
            self.notice_group_id = rawEvent["group_id"]
        elif self.is_group_upload():
            self.notice_user_id = rawEvent["user_id"]
            self.notice_group_id = rawEvent["group_id"]
            self.file = NoticeEvent.File(rawEvent, isGroup=True)
        elif self.is_group_ban():
            self.notice_user_id = rawEvent["user_id"]
            self.notice_group_id = rawEvent["group_id"]
            self.notice_operator_id = rawEvent["operator_id"]
            self.group_ban_type = rawEvent["sub_type"]
            self.ban_time = rawEvent["duration"]
        elif self.is_friend_add():
            self.notice_user_id = rawEvent["user_id"]
        elif self.is_poke():
            self.notice_user_id = rawEvent["target_id"]
            self.notice_operator_id = rawEvent["user_id"]
            if "group_id" in rawEvent.keys():
                self.notice_group_id = rawEvent["group_id"]
        elif self.is_lucky_king():
            self.notice_user_id = rawEvent["target_id"]
            self.notice_group_id = rawEvent["group_id"]
            self.notice_operator_id = rawEvent["user_id"]
        elif self.is_honor():
            self.honor_change_type = rawEvent["honor_type"]
            self.notice_user_id = rawEvent["user_id"]
            self.notice_group_id = rawEvent["group_id"]
        elif self.is_title():
            self.new_title = rawEvent["title"]
            self.notice_user_id = rawEvent["user_id"]
            self.notice_group_id = rawEvent["group_id"]
        elif self.is_group_card():
            self.old_card = rawEvent["card_old"]
            self.new_card = rawEvent["card_new"]
            self.notice_user_id = rawEvent["user_id"]
            self.notice_group_id = rawEvent["group_id"]
        elif self.is_offline_file():
            self.notice_user_id = rawEvent["user_id"]
            self.file = NoticeEvent.File(rawEvent, isGroup=False)
        elif self.is_client_status():
            self.client = NoticeEvent.Client(rawEvent)
        elif self.is_essence():
            self.essence_change_type = rawEvent["sub_type"]
            self.notice_user_id = rawEvent["sender_id"]
            self.notice_group_id = rawEvent["group_id"]
            self.notice_operator_id = rawEvent["operator_id"]
            self.notice_msg_id = rawEvent["message_id"]

    def is_group(self) -> bool:
        """是否是来自群的通知事件"""
        return "group_id" in self.raw.keys()

    def is_group_upload(self) -> bool:
        """是否是群文件上传通知"""
        return self.raw["notice_type"] == "group_upload"

    def is_group_admin(self) -> bool:
        """是否是群管理员变更通知"""
        return self.raw["notice_type"] == "group_admin"

    def is_group_decrease(self) -> bool:
        """是否是群成员减少通知"""
        return self.raw["notice_type"] == "group_decrease"

    def is_group_increase(self) -> bool:
        """是否是群成员增加通知"""
        return self.raw["notice_type"] == "group_increase"

    def is_group_ban(self) -> bool:
        """是否是群禁言通知"""
        return self.raw["notice_type"] == "group_ban"

    def is_friend_add(self) -> bool:
        """是否是好友添加通知"""
        return self.raw["notice_type"] == "friend_add"

    def is_group_recall(self) -> bool:
        """是否是群聊消息撤回通知"""
        return self.raw["notice_type"] == "group_recall"

    def is_friend_recall(self) -> bool:
        """是否是私聊消息撤回通知"""
        return self.raw["notice_type"] == "friend_recall"

    def is_group_card(self) -> bool:
        """是否是群名片变更通知"""
        return self.raw["notice_type"] == "group_card"

    def is_offline_file(self) -> bool:
        """是否是离线文件上传通知（即私聊文件上传）"""
        return self.raw["notice_type"] == "offline_file"

    def is_client_status(self) -> bool:
        """是否是客户端状态通知"""
        return self.raw["notice_type"] == "client_status"

    def is_essence(self) -> bool:
        """是否是精华消息变更通知"""
        return self.raw["notice_type"] == "essence"

    def is_notify(self) -> bool:
        """是否为系统通知（包含群荣誉变更、戳一戳、群红包幸运王、群成员头衔变更）"""
        return self.raw["notice_type"] == "notify"

    def is_honor(self) -> bool:
        """是否是群荣誉变更通知"""
        return self.raw["notice_type"] == "notify" and self.raw["sub_type"] == "honor"

    def is_poke(self) -> bool:
        """是否是戳一戳通知"""
        return self.raw["notice_type"] == "notify" and self.raw["sub_type"] == "poke"

    def is_lucky_king(self) -> bool:
        """是否是群红包幸运王通知"""
        return (
            self.raw["notice_type"] == "notify" and self.raw["sub_type"] == "lucky_king"
        )

    def is_title(self) -> bool:
        """是否是群成员头衔变更通知"""
        return self.raw["notice_type"] == "notify" and self.raw["sub_type"] == "title"

    class File:
        """文件上传通知的文件信息数据结构

        .. admonition:: 提示
           :class: tip

           一般无需手动实例化该类，多数情况会直接使用本类对象，或将本类用作类型注解。
        """

        def __init__(self, rawEvent: dict, isGroup: bool) -> None:
            #: 文件 id，私聊文件上传时为 :obj:`None`
            self.id: Optional[str] = None
            #: 文件名
            self.name: str
            #: 文件大小
            self.size: int
            #: 文件 busid，私聊文件上传时为 :obj:`None`
            self.busid: Optional[int] = None
            #: 文件 url 地址，群聊文件上传时为 :obj:`None`
            self.url: Optional[str] = None

            self.name = rawEvent["file"]["name"]
            self.size = rawEvent["file"]["size"]
            if isGroup:
                self.id = rawEvent["file"]["id"]
                self.busid = rawEvent["file"]["busid"]
            else:
                self.url = rawEvent["file"]["url"]

    class Client:
        """客户端在线状态变更通知的客户端信息数据结构

        .. admonition:: 提示
           :class: tip

           一般无需手动实例化该类，多数情况会直接使用本类对象，或将本类用作类型注解。
        """

        def __init__(self, rawEvent: dict) -> None:
            #: 当前客户端是否在线
            self.online: bool = rawEvent["online"]
            #: 当前客户端 id
            self.id: int = rawEvent["client"]["app_id"]
            #: 客户端设备名称
            self.name: str = rawEvent["client"]["device_name"]
            #: 客户端设备类型
            self.kind: str = rawEvent["client"]["device_kind"]


class MetaEvent(BotEvent):
    """元事件类型

    .. admonition:: 提示
       :class: tip

       一般无需手动实例化该类，多数情况会直接使用本类对象，或将本类用作类型注解。
    """

    def __init__(self, rawEvent: dict) -> None:
        super().__init__(rawEvent)
        #: 收到事件的机器人 qq 号
        self.bot_id: int = rawEvent.get("self_id")  # type: ignore

    @property
    def time(self) -> int:
        """事件发生时刻的时间戳"""
        return self.raw.get("time")  # type: ignore

    @property
    def type(self) -> Literal["meta"]:
        """事件的类型（返回指定字面量）"""
        return "meta"

    def is_lifecycle(self) -> bool:
        """是否是生命周期类型的元事件"""
        return self.raw["meta_event_type"] == "lifecycle"

    def is_heartbeat(self) -> bool:
        """是否是心跳类型的元事件"""
        return self.raw["meta_event_type"] == "heartbeat"


class ResponseEvent(BotEvent):
    """响应事件类型

    .. admonition:: 提示
       :class: tip

       一般无需手动实例化该类，多数情况会直接使用本类对象，或将本类用作类型注解。

    .. admonition:: 提示
       :class: tip

       onebot 标准中未定义“响应”为一种事件类型。但在 melobot 中，“响应”仍然被封装为一种事件类型。
    """

    def __init__(self, rawEvent: dict) -> None:
        super().__init__(rawEvent)

        #: 响应标识符。当响应是对 resp_id 属性不为空的 :class:`.BotAction` 的回应时，此属性才不为 :obj:`None`
        self.id: Optional[str] = None
        #: 响应的状态码
        self.status: int
        #: 响应的数据
        self.data: Optional[dict] = None

        self._init()

    @property
    def time(self) -> int:
        """响应被 melobot 接收时的时间戳"""
        return int(time.time())

    @property
    def type(self) -> Literal["response"]:
        """事件的类型（返回指定字面量）"""
        return "response"

    def _init(self) -> None:
        rawEvent = self.raw
        self.status = rawEvent["retcode"]
        if "echo" in rawEvent.keys() and rawEvent["echo"]:
            self.id = rawEvent["echo"]
        if "data" in rawEvent.keys() and rawEvent["data"]:
            self.data = rawEvent["data"]

    def is_ok(self) -> bool:
        """是否为成功响应"""
        return self.raw["status"] == "ok"

    def is_processing(self) -> bool:
        """是否为被异步处理的响应，即未完成但在处理中"""
        return self.status == 202

    def is_failed(self) -> bool:
        """是否为失败响应"""
        return self.raw["status"] != "ok"
